//
// anyRemote
// a bluetooth remote for your PC.
//
// Copyright (C) 2012-2012 Mikhail Fedotov <anyremote@mail.ru>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

#include "common.h"
#include "executor.h"
#include "utils.h"
#include "alarm.h"

extern char tmp[MAXMAXLEN];

typedef struct {
    char        *file;
    char        *macro;
} type_alarm;


SingleList * _alarmList = NULL;

//////////////////////////////////////////////////////////////////////////////////
//
// Functions related to alarms
//
//////////////////////////////////////////////////////////////////////////////////

// mode: 0 - clear alarms
//       1 - if fired - process, then clear alarms
void manageAlarms(int mode)
{
    struct stat buf;
    
    SingleList* list = _alarmList;
    while (list) {

	type_alarm * v = (type_alarm *) list->data;

        // If file exists
        int fs = stat(v->file, &buf);
        if (fs == 0) {
            if (mode == ALARM_CLEAN) {
                logger(L_DBG,"[ML]: Remove file");
            } else {
                INFO2("[ML]: Alarm %s fired", v->macro);

                eMessage* em = (eMessage*) malloc(sizeof(eMessage));
                em->type  = EM_ALARM;
                em->value = strdup(v->macro);

                sendToExecutor(em);
            }
            remove(v->file);
        }
        list = listSingleNext(list);
    }
}

SingleList * getAlarms() 
{
    return _alarmList;
}

void forgetAlarms()    // just set to NULL, no cleanup
{
    _alarmList = NULL;
}

void addAlarm(char *file, char *macro)
{
    sprintf(tmp,"addAlarm() >%s< >%s<\n", file,macro);
    logger(L_DBG,tmp);

    type_alarm * v = (type_alarm *) calloc(sizeof(type_alarm),1);
    v->file  = strdup(file);
    v->macro = strdup(macro);
    
    _alarmList = listSingleAppend(_alarmList, v);
}

void destroyAlarm(void* ptr) 
{
    type_alarm * v = (type_alarm *) ptr;
    
    //sprintf(tmp,"destroyAlarm() >%s<", v->file);
    //logger(L_DBG,tmp);
    
    free(v->file);
    free(v->macro);
    free(v);
}

void freeAlarms(SingleList* chain)
{
    listSingleFullFree((chain ? chain : _alarmList), destroyAlarm);
    if (!chain) {
        _alarmList = NULL;
    }
}
