//
// anyRemote
// a bluetooth remote for your PC.
//
// Copyright (C) 2006-2012 Mikhail Fedotov <anyremote@mail.ru>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <time.h>
#include <unistd.h>
#ifdef __FreeBSD__
#include <signal.h>
#endif

#include "lib_wrapper.h"
#include "atsend.h"
#include "btio.h"
#include "conf.h"
#include "common.h"
#include "utils.h"
#include "executor.h"
#include "dispatcher.h"
#include "list.h"

#include "pr_frontend.h"
#include "pr_btspp.h"
#include "pr_socket.h"
#ifdef USE_L2CAP
#include "pr_l2cap.h"
#endif
#include "pr_web.h"
#include "pr_stdin.h"
#include "queue.h"
#include "peer.h"

#define TMPDISCONN_TIME 60

extern CONF	conf;
extern char	tmp[MAXMAXLEN];

extern char callerId[MAXLEN];

boolean_t dispIsJoinable = BOOL_YES;

int  callTimer	  = 0;

int  callOn       = 0;

int  gotExitSignal  = 0;

static int  initialized  = 0;

string_t*   readBuffer   = NULL; // store statically, not to reallocate

// Then port will be closed from a forked child use 0
void closePort(int final)
{
    if (final) {
        logger(L_DBG,"[DS]: closePort");
    }

    closePeers(final);

    if (getFrontEnd() > 0) {
        if (final)  {
            sendToFrontEnd("Exiting");
        }
        disconnectFrontEnd();
    }
}

static void handleActiveCall(void)
{
    if (callTimer >= 20) {
        logger(L_DBG, "[DS]: TODO!!!!!! Get call list again");

        // Will set global callerId [MAXLEN];
        /*
        int ret = getClip(callerId);
               if (ret == EXIT_EXACT) {
                   DEBUG2("[DS]: Call still active (%s)", callerId);
               } else {
                   logger(L_DBG,"[DS]: Call was finished");
                   callOn = 0;
                   strcpy(callerId, "NO CALLER ID");

                   sendEventToExecutor(ID_EVT_ENDCALL);
               }
        */
        callTimer = 0;
    } else {
        callTimer++;
    }
}

char* readFromFile(char *cmdTag, char* file, int* size)
{
    FILE *fp;
    struct stat buf;

    long fLen = 0;
    if(stat(file, &buf) == -1) {
        logger(L_ERR,"can't get file size!");
    } else {
        fLen = (int) buf.st_size;

        if (!S_ISREG (buf.st_mode)) {
            logger(L_ERR,"not regular file");
        } else {

            fp=fopen(file,"r");
            if (fp == NULL) {
                logger(L_ERR,"can't open file!");
            } else {

                int prefixSz = strlen(cmdTag);

                uint32_t szh32 = (uint32_t) buf.st_size;
                uint32_t szi32 = htonl(szh32);

                char * fBuffer = (char*) calloc(fLen+prefixSz+7,1);	// 7 =   <size of bin data=4bytes>    ");\0"
                if (fBuffer == NULL) {
                    logger(L_ERR,"no more memory!");

                    *size = -1;
                    return NULL;
                }

                strcat(fBuffer, cmdTag);
                memcpy((void*)fBuffer+prefixSz, (const void *) &szi32, 4);	// length on binary data

                if (fp) {
#ifdef __cplusplus
                    size_t dummy =
#endif
                        fread(fBuffer+prefixSz+4, sizeof(char), fLen, fp);

                    fclose(fp);
                }
                strcpy(fBuffer+prefixSz+fLen+4,");");

                *size = fLen+prefixSz+7;
                return fBuffer;
            }
        }
    }

    if (strncmp(cmdTag, "cover", 5) == 0) {  // cleanup the cover on client

        char * fBuffer = (char*) malloc(12);
        strcpy(fBuffer, "Set(cover);");

        *size = 11;
        return fBuffer;
    }

    *size = -1;
    return NULL;
}

dMessage* getDFinalizer()
{
    dMessage* dm = (dMessage*) malloc(sizeof(dMessage));
    dm->type  = DM_SET;
    dm->size  = 6;
    dm->value = (void*) strdup("End();");
    dm->file  = NULL;
    return dm;
}

void freeDMessage(void *ptr)
{
    dMessage* dm = (dMessage*) ptr;
    if (dm) {
        if (dm->value != NULL) {
            free(dm->value);
        }
        if (dm->file != NULL) {
            free(dm->file);
        }
        free(dm);
    }
}

static int dispatchEventInit(dMessage*  dm)
{
    logger(L_INF, "[DS]: Got init OK event");
    return EXIT_INITOK;
}

static int dispatchEventExit(dMessage* dm)
{
    logger(L_INF, "[DS]: Got exit event");

    gotExitSignal = 1;

    // in server mode wait client to close connection and then exit
    if (disconnectPeers()) {
        logger(L_INF, "[DS]: Got exit event: exit immediately");
        closePort(1);
        return EXIT_ABORT;
    }
    return EXIT_OK;
}

static int dispatchDisconnect(dMessage* dm)
{
    logger(L_INF, "[DS]: Got disconnect event");
    disconnectPeers();
    closePort(0);
    return EXIT_DISCON;
}

static int dispatchEventSendString(dMessage* dm)
{
    DEBUG2("[DS]: Send(string) %s", (char*) dm->value);
    return writePeers(dm);
}

static int dispatchEventSendBytes(dMessage* dm)
{
    DEBUG2("[DS]: Send(bytes) %s", (char*) dm->value);
    return writeBytesPeers((char*) dm->value);
}

static int dispatchEventSet(dMessage* dm)
{
    logger(L_DBG, "[DS]: Set(...)");
    return writePeers(dm);
}

static struct {
    int id;
    int (*hook)(dMessage* p);
} _dispMsgHooks[] = {
    {DM_SET,              dispatchEventSet       },
    {DM_SETFILE,          writeFilePeers         },
    {DM_SENDB,            dispatchEventSendBytes },
    {DM_SENDS,            dispatchEventSendString},
    {DM_CKPD,             writeCKPD              },
    {DM_CMER,             writeCMER              },
    {DM_EVENT_INIT,       dispatchEventInit      },
    {DM_EVENT_DISCONNECT, dispatchDisconnect     },
    {DM_EVENT_EXIT,       dispatchEventExit      }
};

static int checkMsgQueue()
{
    //logger(L_DBG,"[DS]: checkMsgQueue");

    int ret = EXIT_OK;

    // Verify commands from queue (timeout about 1/2 sec)
    dMessage* dm = (dMessage*) queuePop(Q_DISP);
    if (dm != NULL) {

        //DEBUG2("[DS]: Got event %p %d", dm, dm->type);

        if (connected() == EXIT_NOK &&
                !(dm->type == DM_EVENT_INIT || 	// can process these 3 even if no connection
                  dm->type == DM_EVENT_EXIT ||
                  dm->type == DM_EVENT_DISCONNECT)) {

            logger(L_DBG, "[DS]: No connection. Skip event");

        } else {

            ret = (_dispMsgHooks[dm->type].hook)(dm);

        }

        freeDMessage(dm);
    }
    return ret;
}

void parseCommand(char* cmd)
{
    //DEBUG2("[DS]: parseCommand >%s<", cmd);

    char *prev, *next;

    if (cmd == NULL) {
        return ;
    }

    //skip lines starting with \n and \r
    if (cmd[0] == '\r') {
        cmd++;
    }
    if (cmd[0] == '\n') {
        cmd++;
    }

    // most common case
    if (!cmd[0]) {
        return;
    }

    // if recieved multiline command - handle line by line and return
    prev = cmd;
    next = strchr(cmd, '\r');

    if (next == NULL) {	// Java client will send +CKEV: 1,1; +CKEV: 1,0
        next = strchr(cmd, ';');
    }

    /* handle in reader
    if (next == NULL && getIViewer()) {
        next = strchr(cmd, '\3');		// end-of-text marker in CommandFusion
    }*/

    if (next) {
        logger(L_DBG, "[DS]: parseCommand multiline");

        char copy[1024];
        int len;

        do {
            len = next-prev;

            if (len >= 2) {
                memcpy(copy, prev, len);
                copy[len] = 0;

                // use recursion
                parseCommand(copy);
            }
            prev = next+1;
            next = strchr(prev, '\r');
            if (next == NULL) { // Java client will send +CKEV: 1,1; +CKEV: 1,0
                next = strchr(prev, ';');
            }

            /* handle in reader
            if (next == NULL && getIViewer()) {
                next = strchr(prev, '\3');		// end-of-text marker in CommandFusion
            }*/

        } while (next) ;

        // now return
        return;
    }

    logger(L_DBG,"[DS]: -------------------- Command read --------------------");
    DEBUG2("[DS]: parseCommand >%s<", cmd);

    if (IS_OK(cmd)) {   // OK - nothing to do
        return;
    }

    if (strncmp(cmd, "AT+CKPD=", 8) == 0) {        // This is echo of sent message in AT mode; nothing to do
        return;
    }

    if (strncmp(cmd, "Msg:_DEBUG_(", 12) == 0) {   // This is debug message from java client; nothing to do
        return;
    }

    eMessage* em = (eMessage*) malloc(sizeof(eMessage));
    em->type  = EM_KEY;
    em->value = strdup(cmd);

    DEBUG2("[DS]: Send message to executor thread %s", (char*) em->value);
    sendToExecutor(em);

    return;
}

static int read_command(char* buf, int l)
{
    if (!readBuffer) {
        readBuffer = stringNew("");
    }
    stringTruncate(readBuffer,0);

    int rc = readPeers(readBuffer);
    if (rc == EOF) {
        return EOF;
    }

    char* start = readBuffer->str;
    int   sz = rc;
    while (*start == '\r' && sz >= 0) {
        start++;
        sz--;
    }
    strncpy(buf, start, sz);
    buf[sz] = '\0';

    return rc;
}

static void hookInitOnce()
{
    if (initialized == 0) {

        // setgid
        if(conf.uid && getuid()==0) {
            DEBUG2("[DS]: setuid/setgid %d,%d",conf.uid,conf.gid);
            setgid(conf.gid);
            setuid(conf.uid);
        }

        initialized++;
    }
}

int doCheckPassword()
{
    logger(L_DBG,"[DS]: Do password verification");

    char answer[MAXCMDLEN];
    int ret = EXIT_NOK;

    char *pass = getStoredPass();
    if (pass) {
        int gotReply = 1;
        while (1) {

            if (gotReply == 1) {
                sendToFrontEnd("Verifying password");

                dMessage* dm = (dMessage*) malloc(sizeof(dMessage));
                dm->type  = DM_SET;
                dm->size  = 14;
                dm->value = strdup("Get(password);");
                dm->file  = NULL;

                ret = writePeers(dm);

                free(dm->value);
                free(dm);

                if (ret != EXIT_OK) {
                    logger(L_DBG,"[DS]: Fails in send command to retrieve password");
                    continue;
                }

                if (needFinalizer() == EXIT_OK) {
                    dMessage* dm = getDFinalizer();
                    writePeers(dm);
                    free(dm->value);
                    free(dm);
                }

                gotReply = 0;
            }

            int retRead = read_command(answer, sizeof(answer));

            if (retRead == EOF) {
                logger(L_DBG,"[DS]: Got disconnect on verifying password");
                sendToFrontEnd("Disconnected");
                sendDisconnect();
                break;
            } else if (retRead > 0) {
                DEBUG2("[DS]: Password => %s\n",answer);
                //DEBUG2("[DS]: Password => %s\n",answer+16);
                //DEBUG2("[DS]: Compare to => %s\n",pass);

                gotReply = 1;

                if (memcmp(answer,   DEF_MSG, 4) == 0 &&
                        memcmp(answer+4, "_PASSWORD_(,", 12) == 0) {	    // Got verification response from Java client

                    int plen = strlen(pass);
                    if (memcmp(answer+16, pass, plen) == 0) {	// _PASSWORD_(,<password here>)
                        logger(L_INF,"[DS]: Password OK");
                        free(pass);

                        ret = EXIT_OK;
                        break;
                    } else {
                        logger(L_INF,"[DS]: Verification failed. Incorrect password.");
                        sendToFrontEnd("Verification failed");
                    }

                } else {
                    logger(L_INF,"[DS]: Verification failed. Incorrect response.");
                    sendToFrontEnd("Verification failed");
                }
            }

            if (checkMsgQueue() == EXIT_ABORT) {
                return EXIT_ABORT;
            }
            usleep(50000);
        }
    } else {
        ret = EXIT_OK;
    }

    if (pass && ret == EXIT_OK) {
        logger(L_DBG,"[DS]: Password verification OK");
    } else if (pass && ret == EXIT_NOK) {
        logger(L_DBG,"[DS]: Connection problem.");
    } else if (!pass) {
        logger(L_DBG,"[DS]: No password specified. Skip check.");
    }
    return ret;
}

static int doDisconnect()
{
    logger(L_INF, "[DS]: Got disconnected");
    //printf("Got disconnected\n");

    freeBtAddress();

    if (gotExitSignal) {
        logger(L_INF, "[DS]: Got signal, exiting");
        closePort(1);
        return EXIT_ABORT;

    }
    sendToFrontEnd("Disconnected");
    sendDisconnect();
    sendEventToExecutor(ID_EVT_DISCONNECT);

    if (disconnectPeers() == 1) {
        logger(L_INF, "[DS]: Closing the port");
        closePort(0);
    }

    return EXIT_DISCON;
}

void sendFinalizer()
{
    if (needFinalizer() == EXIT_OK) {
        eMessage* em = (eMessage*) malloc(sizeof(eMessage));
        em->type  = EM_AS_IS;
        em->value = strdup("End();");
        sendToExecutor(em);
    }
}

static int doReadPeers()
{
    char answer[MAXCMDLEN];

    int retRead = read_command(answer, sizeof(answer));
    //if (retRead != 0) printf("GOT %d\n",retRead);

    if (retRead == EOF) {
        logger(L_DBG,"[DS]: doReadPeers() EOF");
        return doDisconnect();

    } else if (retRead > 0) {

        parseCommand(answer);
        sendFinalizer();

    }
    return EXIT_OK;
}

int doMessageLoop()
{
    int ret = EXIT_DISCON;

    while (1) {
        //logger(L_DBG,"[DS]: main loop");

        // read from sockets, etc.
        ret = doReadPeers();
        if (ret == EXIT_ABORT) {
            return EXIT_ABORT;
        } else if (ret == EXIT_DISCON) {
            break;
        }

        // Is call still active (timeout about 1 seconds) ?
        if (callOn) {
            handleActiveCall();
        }

        ret = checkMsgQueue();
        if (ret == EXIT_ABORT) {
            return EXIT_ABORT;
        } else if (ret == EXIT_DISCON) {
            break;
        }

        // Main loop timer (1/50 of second)
        usleep(20000);
    } // while (forever)

    return ret;
}

void dispatcherCleanup()
{
    logger(L_DBG,"[DS]: dispatcherCleanup()");
    if (readBuffer) {
        stringFree(readBuffer, BOOL_YES);
    }
    freePeers();
}

pointer_t dispatcherRoutine(pointer_t thread)
{
    int ret = EXIT_OK;

    logger(L_DBG,"[DS]: start dispatcher thread");

    setCmdFiles();

    // wait init ok event
    while (1) {
        ret = checkMsgQueue();
        if (ret == EXIT_ABORT) {
            dispatcherCleanup();
            return NULL;
        } else if (ret == EXIT_INITOK) {
            break;
        }
        logger(L_DBG,"[DS]: wait init OK event");
        usleep(50000);
    }

    int dmn = autoConnect();
    int rs  = getRetrySecs();
    strcpy(callerId, "NO CALLER ID");

    if (definePeers() == EXIT_ABORT) {
        logger(L_DBG,"[DS]: Incorrect device specification");
        dispatcherCleanup();
        sendAbort();
        return NULL;
    }

    while (1) {

        logger(L_DBG,"[DS]: ************ outer loop **********");

        if(openPeers() == EXIT_OK) {   // Open device

            logger(L_DBG,"[DS]: Device open OK");

            hookInitOnce();

            dispIsJoinable = BOOL_NO;

            if (acceptPeers() == 1) {  // Init modem: AT, ATE0, AT+CMER, in server mode waits for connection

                logger(L_DBG,"[DS]: Init connection OK");

                dispIsJoinable = BOOL_YES;

                int pass_ok = (usePassword() ? doCheckPassword() : EXIT_OK);

                if (pass_ok == EXIT_ABORT) {
                    dispatcherCleanup();
                    return NULL;
                }

                if (pass_ok == EXIT_OK) {  // not EOF

                    sendToFrontEnd("Connected");
                    sendEventToExecutor(ID_EVT_CONNECT);

                    logger(L_INF,"[DS]: Start event forwarding");

                    int retLoop = doMessageLoop();
                    if (retLoop == EXIT_ABORT) {
                        dispatcherCleanup();
                        return NULL;
                    }
                    // EXIT_DISCON / EXIT_NOK is OK here

                } else {  	// not EOF after password verification
                    logger(L_DBG,"[DS]: Connection EOF");
                }
            } else {		// init port
                logger(L_DBG,"[DS]: Init connection error");
            }
        } else {		// open port
            logger(L_DBG,"[DS]: Device open error");
        }

        //printf("Connection closed or lost\n");
        logger(L_INF, "[DS]: Connection closed or lost");

        // Can't open port or it closed again

        int isServer = isServerMode();

        if (!gotExitSignal &&
                (dmn || isServer == EXIT_OK || ret == EXIT_DISCON)) {

            int timeout;

            if (isServer == EXIT_OK) {

                timeout = 2;	// wait only 2 seconds

            } else if (ret == EXIT_DISCON) {

                timeout = TMPDISCONN_TIME;
                ret = EXIT_OK;

            } else {
                timeout = rs;
            }

            INFO2("[DS]: Wait %d seconds to connect/open server socket ...", timeout);
            //printf("Wait %d seconds to connect/open server socket ...\n", timeout);

            dispIsJoinable = BOOL_NO;
            sleep(timeout);
            dispIsJoinable = BOOL_YES;

        } else {	// Proceed to exit
            printf("Exiting ...\n");
            sendAbort();
            break;
        }
    }

    // Finish all
    logger(L_INF, "[DS]: Stop dispatcher ...");
    closePort(1);
    dispatcherCleanup();

    return NULL;
}

void sendToDispatcher(dMessage *buf)
{
    if (queuePush(Q_DISP, (void*) buf) == RC_OK) {
        DEBUG2("send to dispatcher %d", buf->type);
    }
}
