//
// anyRemote
// a bluetooth remote for your PC.
//
// Copyright (C) 2012-2012 Mikhail Fedotov <anyremote@mail.ru>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// Main loop
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>

#include "lib_wrapper.h"

#include "conf.h"
#include "utils.h"
#include "alarm.h"
#include "queue.h"
#include "dispatcher.h"
#include "peer.h"
#include "pr_frontend.h"

extern void aboutToExit();

boolean_t adminRoutine(pointer_t data)
{
    int *mi = (int*) queuePop(Q_MAIN);

    if (mi != NULL) {

        int m = *mi;
        free(mi);

        if (m == M_ABORT) {
            logger(L_DBG,"[ML]: got exit request");
            aboutToExit();	// exit here

        } else if (m == M_DISCONNECT) {
            logger(L_DBG,"[ML]: got disconnect request");

            if (!autoConnect() && 
	        needExit() == EXIT_OK) {   // no active connection (and all peers are clients)
                logger(L_DBG,"[ML]: Exit because of disconnect");
                aboutToExit();	// exit here
            }
        }
    }
    return BOOL_YES;
}

#ifdef USE_GLIB

static GMainLoop *loop 	 = NULL;

gboolean alarmTimeout(gpointer data)
{
    //logger(L_DBG,"[ML]: alarmTimeout");
    manageAlarms(ALARM_CHECK);
    return TRUE;
}

void loopStart()
{
    loop = g_main_loop_new(NULL, FALSE);
    
    if (getFrontEnd() > 0) {
        g_timeout_add(200, frontendRoutine, NULL); 	// 1/5 of second
    } else {
        logger(L_DBG,"Do not setup front-end connection");
    }

    g_timeout_add(100,   adminRoutine, NULL); 		// 1/10 of second
    g_timeout_add(15000, alarmTimeout, NULL);

    //logger(L_DBG,"g_main_loop_run");
    g_main_loop_run(loop);
}

void loopStop()
{
    g_main_loop_quit(loop);
}

void loopDestroy()
{
    g_main_loop_unref(loop);
}

#else

int  alarmTimer    = 0;
//int  adminTimer    = 0;
int  frontendTimer = 0;

void loopStart()
{
    while (1) {      
    	
	/* about 1/10 of second
    	if (adminTimer == 2) {
            adminRoutine(NULL);
    	    adminTimer = 0;
    	} else {
    	    adminTimer++;
	    
    	}*/
	
	if (queueCanPop(Q_MAIN)) {
	    adminRoutine(NULL);
	}
	
        // about 1/5 of second
   	if (frontendTimer == 4) {
            frontendRoutine(NULL);
    	    frontendTimer = 0;
    	} else {
    	    frontendTimer++;
    	}
	
	// about 15 seconds
  	if (alarmTimer == 300) {
    	    manageAlarms(ALARM_CHECK);
    	    alarmTimer = 0;
    	} else {
    	    alarmTimer++;
    	}
    	
    	// Main loop timer (1/20 of second)
    	usleep(50000);
    }  
}


void loopStop() 	// do nothing
{
}

void loopDestroy() 	// do nothing
{
}

#endif
