//
// anyRemote
// a bluetooth remote for your PC.
//
// Copyright (C) 2011-2012 Mikhail Fedotov <anyremote@mail.ru>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <dirent.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>

#define SERVER     "webserver/1.1"
#define PROTOCOL   "HTTP/1.1"
#define RFC1123FMT "%a, %d %b %Y %H:%M:%S GMT"

#include "str.h"
#include "utils.h"
#include "conf.h"
#include "cmds.h"
#include "queue.h"
#include "mutex.h"
#include "thread.h"
#include "list.h"
#include "var.h"
#include "pr_web.h"

extern char tmp[MAXMAXLEN];
extern int  gotExitSignal;

int visits  =  0;                  // counts client connections
int clients =  0;                  // counts client connections
int runMode = SERVER_WEB;  // or SERVER_CMXML

int port = 5050;
int sock = -1;
long cookie = 0;
int secure = NO_COOKIE;
int curForm;

boolean_t runWeb = BOOL_YES;
boolean_t answerReady = BOOL_NO;
boolean_t askPassword = BOOL_NO;
boolean_t bottomlineSkin = BOOL_NO;

char* confDir = NULL;

int screenSize = 240;  // min(width,heigth)
int refreshPage = -1;  // refrech time in seconds, -1 meand no refresh

int      _xmlScreenW = 298;    // Cisco-7941 as default
int      _xmlScreenH = 144;    //
int      _xmlLayoutW = 298;    //
int      _xmlLayoutH = 144;    //
boolean_t _xmlScreenG = BOOL_YES;   //
boolean_t _xmlLayoutOk = BOOL_NO;

// Global data

// Predefined templates
#define HTTP_HEAD1 "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\n<html>\n<head>\n<title>"
#define HTTP_HEAD2 "</title>\n"
#define HTTP_HEAD3 "<meta "
#define HTTP_HEAD4 " content=\"text/html; charset=utf-8\">\n"
#define HTTP_HEAD5 "</head>\n<body"
#define HTTP_HEAD6 "><center>"
#define HTTP_SCRIPT      "<script>\nfunction clearForms()\n{\n  var i;\n  for (i = 0; (i < document.forms.length); i++) {\n    document.forms[i].reset();\n  }\n}\n</script>\n"
#define HTTP_SCRIPT_CALL " onload=\"clearForms()\""

#define HTTP_HEAD_CSS_0  "<style type=\"text/css\">\n"
#define HTTP_HEAD_CSS_1  "table{\n\ttext-align:left;\n}\n"
#define HTTP_HEAD_CSS_B1 "body {\n\tcolor:#"
#define HTTP_HEAD_CSS_B2 "; \n\tbackground-color:#"
#define HTTP_HEAD_CSS_B3 ";\n}\n"
#define HTTP_HEAD_CSS_L1 "select {\n\tcolor:#"
#define HTTP_HEAD_CSS_L3 "; \n\twidth:"
#define HTTP_HEAD_CSS_BT1 "button {\n\tcolor:#"

#define HTTP_HEAD_CSS_99 "</style>\n"

#define HTTP_TABLE1 "\n<table border=\"1\" cellpadding=\""
#define HTTP_TABLE2 "\" cellspacing=\"2\">\n<tbody>\n<tr>\n"
//#define HTTP_TABLE3 "	<td style=\"vertical-align: top; text-align: center; width=5%;\">"
#define HTTP_TABLE3 "	<td>"

#define HTTP_MENU1 " [<a href=\""
#define HTTP_MENU2 ".menu\">"
#define HTTP_MENU3 "</a>] "

#define HTTP_LIST1 "<select name=\"list\" size=15>\n"

#define HTTP_FORM  "\n<form action=\"/\" method=\"post\">\n"

#define HTTP_ACTION    "action="
#define HTTP_EDITFIELD "http_editfield="

#define XML_SOFTKEY_NUM 4

#define XML_HEAD1  "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"

//#define XML_CFORM1  "<CiscoIPPhoneMenu>\n"
//#define XML_CFORM99 "</CiscoIPPhoneMenu>\n"
#define XML_CFORM1  "<CiscoIPPhoneGraphicFileMenu>\n"
#define XML_CFORM99 "</CiscoIPPhoneGraphicFileMenu>\n"

/*
x-CiscoIPPhoneModelName: CP-7941G
x-CiscoIPPhoneSDKVersion: 7.0.1
x-CiscoIPPhoneDisplay: 298,144,3,G

Cisco Unified IP Phone  7920      128 x 59 Grayscale 1
Cisco Unified IP Phones 7940G
                        7960G     133 x 65 Grayscale 2
Cisco Unified IP Phones 7941G,
                        7941G-GE,
                        7961G,
			7961G-GE  298 x 144 Grayscale 4

Cisco Unified IP Phone  7970G
                        7971G     298 x 168 Color     12
*/

#define XML_LIST1  "<CiscoIPPhoneMenu>\n"
#define XML_LIST99 "</CiscoIPPhoneMenu>\n"

#define XML_TEXT1  "<CiscoIPPhoneText>\n"
#define XML_TEXT99 "</CiscoIPPhoneText>\n"

#define XML_EFORM1  "<CiscoIPPhoneInput>\n"
#define XML_EFORM99 "</CiscoIPPhoneInput>\n"

#define XML_WMAN1  "<CiscoIPPhoneImageFile>\n"
#define XML_WMAN99 "</CiscoIPPhoneImageFile>\n"

#define HTTP_EDIT1 "<h4>"
#define HTTP_EDIT2 "</h4></center>\n<form action=\"\" method=\"post\">\n<input type=\"text\" name=\"http_editfield\" value=\""
#define HTTP_EDIT3 "</h4></center>\n<form action=\"\" method=\"post\">\n<input type=\"password\" name=\"http_editfield\" value=\""
#define HTTP_EDIT4 "\" />\n<hr>"

#define XML_BUTTON_PRESS  "xml_button="
#define XML_SHORT_MENU    "xml_me="
#define XML_LONG_MENU     "xml_mw="
#define XML_LIST_MENU     "xml_lm="
#define XML_LIST_MENU2    "xml_ll="
#define XML_LIST_MENU_EXT "xml_lext"
#define XML_EFIELD_CANCEL "xml_ec"
#define XML_EFIELD_SUBMIT ";?xml_ef="

#define XML_MAX_LIST_SIZE 99

#define IVIEW_HEAD1 "<gui>\n  <properties>\n    <project>anyRemote GUI</project>\n    <designer>anyRemote</designer>\n    <controlsystem>\n      <hostname>"
#define IVIEW_HEAD2 "</hostname>\n      <commandPort>"
#define IVIEW_HEAD3 "</commandPort>\n    </controlsystem>\n    <size>\n      <portrait width=\"320\" height=\"480\" />\n      <landscape width=\"480\" height=\"320\" />\n    </size>\n    <imagefolder>\n    </imagefolder>\n    <cache images=\"0\" />\n    <debug loaderrors=\"0\" />\n  </properties>"
#define IVIEW_TAIL  "</gui>"

string_t* httpTail;

// control form
string_t* cf_caption = NULL;
string_t* cf_title   = NULL;
string_t* cf_status  = NULL;
string_t* cf_volume  = NULL;
boolean_t useVolume   = BOOL_NO;
string_t* cf_font    = NULL;
string_t* cf_fg      = NULL;
string_t* cf_bg      = NULL;
string_t* cf_icons[12] = {NULL};
string_t* cf_cover   = NULL;
string_t* window     = NULL;

int iconPadding   = 0;
int iconSize      = 64;
int iconSizeScale = 64;

// list form
string_t* ls_caption = NULL;
string_t* ls_font    = NULL;
string_t* ls_fg      = NULL;
string_t* ls_bg      = NULL;
SingleList *ls_items = NULL;
int ls_selidx = 1;

// text form
string_t* tx_caption = NULL;
string_t* tx_font    = NULL;
string_t* tx_fg      = NULL;
string_t* tx_bg      = NULL;
string_t* tx_text    = NULL;

// edit field
string_t* ef_caption = NULL;
string_t* ef_text    = NULL;
string_t* ef_label   = NULL;

SingleList * user_menu = NULL;

SingleList * clientSockets = NULL;

string_t* serverIP = NULL;

static const char* i2string(int data,char* buf)
{
    sprintf(buf,"%d",data);
    return buf;
}

static void stringItemFree(pointer_t data)
{
    stringFree(data,(((string_t*) data)->str != NULL));
}

static void listItemFree(pointer_t data)
{
    listItem* item = (listItem*) data;
    if (item->icon) {
        stringFree(item->icon,   (item->icon->str   != NULL));
    }
    if (item->string) {
        stringFree(item->string, (item->string->str != NULL));
    }
    free(item);
    item = NULL;
}

void freeWMessage(void *ptr)
{
    wMessage *wm = (wMessage *) ptr;
    if (wm->string) {
        stringFree(wm->string,  BOOL_YES);
        wm->string = NULL;
    }
    free(wm);
}

static void sendToWebServer(wMessage *buf)
{
    if (queueExists(Q_WEB) == RC_OK && buf) {
        mutexLock(M_WEB);
        DEBUG2("send to web server %d %s", buf->button, (buf->string ? buf->string->str : "no text"));
        queuePush(Q_WEB, buf);
        mutexUnlock(M_WEB);
    }
}

static const char *get_mime_type(const char *name)
{
    char *ext = strrchr(name, '.');
    if (!ext) {
        return NULL;
    }

    if (strcmp(ext, ".html") == 0 || strcmp(ext, ".htm") == 0) {
        return "text/html";
    }
    if (strcmp(ext, ".jpg") == 0 || strcmp(ext, ".jpeg") == 0) {
        return "image/jpeg";
    }
    if (strcmp(ext, ".gif") == 0) {
        return "image/gif";
    }
    if (strcmp(ext, ".png") == 0) {
        return "image/png";
    }
    if (strcmp(ext, ".css") == 0) {
        return "text/css";
    }
    if (strcmp(ext, ".au") == 0) {
        return "audio/basic";
    }
    if (strcmp(ext, ".wav") == 0) {
        return "audio/wav";
    }
    if (strcmp(ext, ".avi") == 0) {
        return "video/x-msvideo";
    }
    if (strcmp(ext, ".mpeg") == 0 ||  strcmp(ext, ".mpg") == 0) {
        return "video/mpeg";
    }
    if (strcmp(ext, ".mp3") == 0) {
        return "audio/mpeg";
    }
    return NULL;
}

static int sendData(int fd, const char *s)
{
	//printf("%s\n",s);
    int bytes_sent = send(fd,s,strlen(s),MSG_NOSIGNAL);
    if (bytes_sent != strlen(s)) {
        ERROR2("[WS]: Error on send data: sent %d from %d", bytes_sent, (int) strlen(s));
        return -1;
    }
    return 0;
}

void sendCookie(int fd)
{
    if (secure == NO_COOKIE) {
        return;
    }

    char b[32];
    sendData(fd, "Set-Cookie: anyremote_id=");
    cookie = random();
    sprintf(b,"%ld",cookie);
    sendData(fd, b);
    sendData(fd, "\r\n");

    INFO2("[WS]: sendCookie %s",b);
    secure = COOKIE_SENT;
}

static void sendHeaders(int fd, int status,
                        const char *title, const char *extra, const char *mime,
                        int length, time_t date, int cookie)
{
    char f[4096];
    //INFO2("[WS]: sendHeaders %s",title);

    sprintf(f, "%s %d %s\r\n", PROTOCOL, status, title);
    sendData(fd, f);

    sprintf(f, "Server: %s\r\n", SERVER);
    sendData(fd, f);

    time_t now = time(NULL);
    char timebuf[128];
    strftime(timebuf, sizeof(timebuf), RFC1123FMT, gmtime(&now));

    sprintf(f, "Date: %s\r\n", timebuf);
    sendData(fd, f);

    if (extra) {
        sprintf(f, "%s\r\n", extra);
        sendData(fd, f);
    }
    if (mime) {
        sprintf(f, "Content-Type: %s\r\n", mime);
        sendData(fd, f);
    }

    sprintf(f, "Cache-Control: public, max-age=36000\r\n");
    sendData(fd, f);

    if (length >= 0) {
        sprintf(f, "Content-Length: %d\r\n", length);
        sendData(fd, f);
    }
    sendCookie(fd);

    if (date != -1) {
        strftime(timebuf, sizeof(timebuf), RFC1123FMT, gmtime(&date));
        sprintf(f, "Last-Modified: %s\r\n", timebuf);
        sendData(fd, f);
    }

    sprintf(f, "Connection: close\r\n\r\n");
    sendData(fd, f);
}

static void sendError(int fd, int status, char *title, char *extra, char *text)
{
    char f[4096];

    INFO2("[WS]: sendError %d", status);
    sendHeaders(fd, status, title, extra, "text/html", -1, -1, 0);

    sprintf(f, "<HTML><HEAD><TITLE>%d %s</TITLE></HEAD>\r\n", status, title);
    sendData(fd, f);

    if (status != 304) { // 304  - Not Modified

        sprintf(f, "<BODY><H4>%d %s</H4>\r\n", status, title);
        sendData(fd, f);

        sprintf(f, "%s\r\n", text);
        sendData(fd, f);

        sprintf(f, "</BODY></HTML>");
        sendData(fd, f);

    } else {
        time_t now = time(NULL);
        char timebuf[128];
        strftime(timebuf, sizeof(timebuf), RFC1123FMT, gmtime(&now));

        sprintf(f, "Date: %s\r\n", timebuf);
        sendData(fd, f);
    }


    sprintf(f, "\r\n");
    sendData(fd, f);
}

static void sendFile(int fd, char *path, struct stat *statbuf)
{
    char data[4096];
    int n,bytes_sent;

    INFO2("[WS]: sendFile %s",path);

    FILE *file = fopen(path, "r");
    if (!file) {
        ERROR2("[WS]: Access denied: %s", path);
        sendError(fd, 403, "Forbidden", NULL, "Access denied.");

    } else {

        int length = S_ISREG(statbuf->st_mode) ? statbuf->st_size : -1;
        sendHeaders(fd, 200, "OK", NULL, get_mime_type(path), length, statbuf->st_mtime, 0);

        while ((n = fread(data, 1, sizeof(data), file)) > 0) {
            //INFO2("read %d bytes from file",n);

            bytes_sent = send(fd,data,n,0);
            if (n != bytes_sent) {
                ERROR2("[WS]: Error on send file %s", path);
                break;
            }
        }

        fclose(file);
    }
    sendData(fd, "\r\n");
}

static void sendGeneratedImage(int fd, char *path)
{
    INFO2("[WS]: sendGeneratedImage %s", path);

    const char* home = getenv("HOME");
    if (!home) {
        return;
    }

    struct stat statbuf;

    string_t* img = stringNew(home);
    stringAppend(img,"/.anyRemote/");

    stringAppend(img,path);

    INFO2("[WS]: sendGeneratedImage full name %s", img->str);

    if (stat(img->str, &statbuf) >= 0) {
        sendFile(fd, img->str, &statbuf);
    } else {
        sendError(fd, 404, "Not Found", NULL, "File not found.");
    }
    stringFree(img, BOOL_YES);
}

static void sendIcon(int fd, char *path)
{
    INFO2("[WS]: sendIcon %s", path);

    struct stat statbuf;

    string_t* icon = stringNew(confDir ? confDir : ".");
    if (getIViewer()) {
   	    stringAppend(icon,"/Utils/iViewer");
	    stringAppend(icon,path);
    } else {
        stringAppend(icon,"/Icons/");

        char b[32];
        stringAppend(icon,i2string(iconSize,b));
        if (path[0] != '/') {
            stringAppend(icon,"/");
        }
        stringAppend(icon,path);
    }

    INFO2("[WS]: sendIcon full name %s", icon->str);

    if (stat(icon->str, &statbuf) >= 0) {
        sendFile(fd, icon->str, &statbuf);
    } else {
        sendError(fd, 404, "Not Found", NULL, "File not found.");
    }
    stringFree(icon, BOOL_YES);
}

static void sendFavicon(int fd)
{
    struct stat statbuf;

    string_t* icon = stringNew(confDir ? confDir : ".");
    stringAppend(icon,"/Icons/anyRemote.png");

    if (stat(icon->str, &statbuf) >= 0) {
        sendFile(fd, icon->str, &statbuf);
    } else {
        //ERROR2("[WS]: sendFavicon can not open favicon file");
	sendError(fd, 404, "Not Found", NULL, "File not found.");
    }
    stringFree(icon, BOOL_YES);
}

static void sendIViewerGui(int fd) 
{
	INFO2("[WS]: sendIViewerGui");
	
    sendData(fd, IVIEW_HEAD1);
    
    if (!serverIP) {
	    ERROR2("[WS]: sendIViewerGui can not determine TCP address");
		return;
	}
    sendData(fd, serverIP->str);
    
    sendData(fd, IVIEW_HEAD2);
     
    int tcpPort = getIViewerTcpPort();
    if (tcpPort < 0) {
    	ERROR2("[WS]: sendIViewerGui can not determine TCP port");
    	return;
    }
    char num[16];
    sprintf(num,"%d", tcpPort);
    
    sendData(fd, num);
    sendData(fd, IVIEW_HEAD3);

    string_t* guiFile = stringNew(confDir ? confDir : ".");
    stringAppend(guiFile,"/Utils/anyremote.gui");
    
    struct stat statbuf;
    if (stat(guiFile->str, &statbuf) >= 0) {
        sendFile(fd, guiFile->str, &statbuf);
    } else {
        ERROR2("[WS]: sendIViewerGui can not open GUI file");
    }
    sendData(fd, IVIEW_TAIL);
}

char abuf[8];

static char* getAccessKey(int index)
{
    if (index == 10) {
        abuf[0] = '*';
        abuf[1] = '\0';
    } else if (index == 11) {
        abuf[0] = '0';
        abuf[1] = '\0';
    } else if (index == 12) {
        abuf[0] = '#';
        abuf[1] = '\0';
    } else {
        sprintf(abuf,"%d",index);
    }
    return abuf;
}

static void addLink(string_t* addTo, boolean_t trailSlash)
{
    stringAppend(addTo, "http://");
    stringAppend(addTo, serverIP->str);
    stringAppend(addTo, ":");

    char num[16];
    sprintf(num,"%d", port);

    stringAppend(addTo, num);
    if (trailSlash) {
        stringAppend(addTo, "/");
    }
}

void sendForm(int fd, string_t* httpPageH, string_t* httpPageT)
{
    string_t* header = stringNew("HTTP/1.1 200 OK\r\n");

    if (runMode == SERVER_CMXML) {
        stringAppend(header, "Content-type: text/xml; charset=UTF-8\r\nConnection: close\r\n");
    } else {
        stringAppend(header, "Content-type: text/html\r\nCache-Control: no-cache, must-revalidate\r\n");
    }

    if (refreshPage > 0) {
        stringAppend(header, "Refresh: ");

        char num[16];
        sprintf(num,"%d", refreshPage);
        stringAppend(header, num);

        stringAppend(header, "; url=");
        addLink(header, BOOL_YES);

        stringAppend(header, "\r\n");
    } else {
        if (runMode == SERVER_CMXML) {
            stringAppend(header, "Expires: -1\r\n");
        }
    }

    //printf("HEADER: %s",header->str);
    sendData(fd, header->str);

    if (runMode == SERVER_CMXML) {
        sendCookie(fd);
    }

    sendData(fd, "\r\n");

    sendData(fd, httpPageH->str);
    //printf("%s",httpPageH->str);

    /*if refresh < 0:
    	rStr = "%s" % (uploadTmout)
    	string_t* httpRate = stringNew('HTTP-EQUIV=REFRESH CONTENT=\"' + rStr + ';URL=/\"');
    sendData(fd, httpRate->str);
    stringFree(httpRate,  BOOL_YES);
    */

    sendData(fd, httpPageT->str);
    //printf("%s",httpPageT->str);
}

static void addDefMenu()
{
    INFO2("[WS]: addDefMenu %d", curForm);
    switch (curForm) {

    case TX:
    case LI:
        user_menu = listSingleAppend(user_menu, stringNew("Back"));
        break;

    case EF:
        user_menu = listSingleAppend(user_menu, stringNew("Back"));
        user_menu = listSingleAppend(user_menu, stringNew("Ok"));
        break;

    case CF:
        //user_menu = listSingleAppend(user_menu, stringNew("Disconnect"));
        //user_menu = listSingleAppend(user_menu, stringNew("Exit"));
        break;

    case WM:
        break;
    }
}

static void freeList()
{
    listSingleFullFree(ls_items, listItemFree);
    ls_items = NULL;
    ls_selidx = 1;
}

static void freeMenu()
{
    INFO2("[WS]: freeMenu %d", curForm);
    listSingleFullFree(user_menu, stringItemFree);
    user_menu = NULL;
}

static void setupDefMenu()
{
    INFO2("[WS]: setupDefMenu %d", curForm);
    freeMenu();
    addDefMenu();
}

static void addFormMenu(int formId, string_t* page)
{
    SingleList* list = user_menu;

    INFO2("[WS]: addFormMenu %d #%d", curForm, listSingleLength(list));
    switch (formId) {

    case CF:
    case TX:
    case WM:
        if (listSingleLength(list) > 0) {

            string_t* menu = stringNew("<HR>");

            //stringAppend(menu, "<div style=\"width:240px; overflow:auto;\">");

            while(list) {

                //INFO2("[WS]: addFormMenu add item ");
                if (list->data && ((string_t*) list->data)->str) {

                    stringAppend(menu, HTTP_MENU1);
                    stringAppend(menu, ((string_t*) list->data)->str);
                    stringAppend(menu, HTTP_MENU2);
                    stringAppend(menu, ((string_t*) list->data)->str);
                    stringAppend(menu, HTTP_MENU3);

                    //INFO2("[WS]: addFormMenu item %s", ((string_t*) list->data)->str);
                }
                list = listSingleNext(list);
            }
            //stringAppend(menu, "</div>");
            stringAppend(menu, "<HR>");
            stringAppend(page, menu->str);
        }
        break;

    case LI:
    case EF:

        if (listSingleLength(list) > 0) {

            string_t* menu = stringNew("");

            int i = 0;
            while(list) {

                stringAppend(menu, "<input name=\"action\" type=\"submit\" value=\"");
                stringAppend(menu, ((string_t*) list->data)->str);
                stringAppend(menu, "\">\n");

                i++;
                list = listSingleNext(list);
            }
            stringAppend(menu, "<HR>");
            stringAppend(page, menu->str);
        }
        break;
    }
}

static void addXMLFormMenu(int formId, string_t* page)
{
    SingleList* list = user_menu;

    INFO2("[WS]: addXMLFormMenu %d #%d", curForm, listSingleLength(list));
    switch (formId) {

    case CF:
    case TX:
    case WM: {
        int llen = listSingleLength(list);

        if (llen > 0) {


            boolean_t longMenu = (llen > XML_SOFTKEY_NUM ? BOOL_YES : BOOL_NO);

            int position = 1;

            while(list) {

                if (longMenu && position == XML_SOFTKEY_NUM) {

                    stringAppend(page, "  <SoftKeyItem>\n    <Name>>></Name>\n    <URL>");
                    addLink(page, BOOL_YES);
                    stringAppend(page, XML_LONG_MENU);

                    char buf[8];
                    sprintf(buf,"%d", formId);
                    stringAppend(page, buf);

                    sprintf(buf,"%d", position);
                    stringAppend(page, "</URL>\n    <Position>");
                    stringAppend(page, buf);
                    stringAppend(page, "</Position>\n  </SoftKeyItem>\n");

                    break;
                }

                if (list->data && ((string_t*) list->data)->str) {

                    INFO2("[WS]: addXMLFormMenu item %s",((string_t*) list->data)->str);

                    stringAppend(page, "  <SoftKeyItem>\n    <Name>");
                    stringAppend(page, ((string_t*) list->data)->str);
                    stringAppend(page, "</Name>\n    <URL>");

                    addLink(page, BOOL_YES);

                    char buf[8];
                    sprintf(buf,"%d", position);

                    stringAppend(page, XML_SHORT_MENU);
                    stringAppend(page, buf);

                    stringAppend(page, "</URL>\n    <Position>");
                    stringAppend(page, buf);
                    stringAppend(page, "</Position>\n  </SoftKeyItem>\n");
                }
                list = listSingleNext(list);

                position++;
                if (!longMenu && position > XML_SOFTKEY_NUM) {  // should not happen
                    break;
                }
            }
        }
        break;
    }


    //case LI:  -- sendXMLMenu used instead
    /*case EF: -- use custom code

            if (listSingleLength(list) > 0) {

            int position = 1;

    	while(list) {

    	    if (list->data && ((string_t*) list->data)->str) {

    		stringAppend(page, "  <SoftKeyItem>\n    <Name>");
    		stringAppend(page, ((string_t*) list->data)->str);
    		stringAppend(page, "</Name>\n    <URL>");

    		addLink(page, BOOL_YES);

    		char buf[8];
    		sprintf(buf,"%d", position);

    		stringAppend(page, XML_SHORT_MENU);
    		stringAppend(page, buf);

    		stringAppend(page, "</URL>    <Position>");
    		stringAppend(page, buf);

    		stringAppend(page, "</Position>\n  </SoftKeyItem>\n");
                    }
    	    list = listSingleNext(list);

    	    position++;
    	    if (position > XML_SOFTKEY_NUM) {
    	        break;
    	    }
                }
        }
        break;*/
    }
}

static string_t* renderFormHead(int form, string_t* caption, string_t* fg, string_t* bg)
{
    string_t* head = stringNew(HTTP_HEAD1);
    stringAppend(head, caption->str);
    stringAppend(head, HTTP_HEAD2);
    if (form == LI) {
        stringAppend(head, HTTP_SCRIPT);
    }
    stringAppend(head, HTTP_HEAD3);
    stringAppend(head, HTTP_HEAD4);

    // Do some CSS styling
    stringAppend(head, HTTP_HEAD_CSS_0);
    //stringAppend(head, HTTP_HEAD_CSS_1);

    stringAppend(head, HTTP_HEAD_CSS_B1);
    stringAppend(head, fg->str);
    stringAppend(head, HTTP_HEAD_CSS_B2);
    stringAppend(head, bg->str);
    stringAppend(head, HTTP_HEAD_CSS_B3);

    if (form == CF) {
        stringAppend(head, HTTP_HEAD_CSS_BT1);
        stringAppend(head, fg->str);
        stringAppend(head, HTTP_HEAD_CSS_B2);
        stringAppend(head, bg->str);
        stringAppend(head, ";\n\tborder-style:none;\n}\n");
    }

    if (form == LI) {
        stringAppend(head, HTTP_HEAD_CSS_L1);
        stringAppend(head, fg->str);
        stringAppend(head, HTTP_HEAD_CSS_B2);
        stringAppend(head, bg->str);
        stringAppend(head, HTTP_HEAD_CSS_L3);
        char b[32];
        stringAppend(head, i2string(screenSize,b));
        stringAppend(head, "px");

        stringAppend(head, HTTP_HEAD_CSS_B3);
    }
    stringAppend(head, HTTP_HEAD_CSS_99);

    stringAppend(head, HTTP_HEAD5);
    if (form == LI) {
        stringAppend(head, HTTP_SCRIPT_CALL);
    }
    stringAppend(head, HTTP_HEAD6);

    return head;
}

static string_t* renderXMLFormHead(int form, string_t* caption)
{
    string_t* head = stringNew(XML_HEAD1);

    if (form == CF) {
        stringAppend(head,XML_CFORM1);
    } else if (form == LI) {
        stringAppend(head,XML_LIST1);
    } else if (form == TX) {
        stringAppend(head,XML_TEXT1);
    } else if (form == EF) {
        stringAppend(head,XML_EFORM1);
    } else if (form == WM) {
        stringAppend(head,XML_WMAN1);
    }

    stringAppend(head, "  <Title>");
    stringAppend(head, caption->str);
    stringAppend(head, "</Title>\n");
    stringAppend(head, "  <Prompt>");
    if (form == CF) {
        stringAppend(head, cf_title->str);
    } else {
        stringAppend(head, caption->str);
    }
    stringAppend(head, "</Prompt>\n");

    return head;
}

// Send menu as list (CMXML)
static void sendXMLMenu(int fd, int form, int idx)
{

    string_t* caption = stringNew("Menu");
    string_t* httpPageH = renderXMLFormHead(LI,caption);  // menu-as-list
    stringFree(caption, BOOL_YES);

    string_t* page = stringNew("");

    SingleList* list = user_menu;

    INFO2("[WS]: sendXMLMenu %d #%d", form, listSingleLength(list));
    if (listSingleLength(list) > 0) {

        int position = 1;

        while(list) {
            if (list->data && ((string_t*) list->data)->str) {

                stringAppend(page, "  <MenuItem>\n   <Name>");

                if (((string_t*) list->data)->len > CMXML_MAX_STRING_SIZE) {
                    stringAppendLen(page, ((string_t*) list->data)->str, CMXML_MAX_STRING_SIZE);
                } else {
                    stringAppend(page, ((string_t*) list->data)->str);
                }
                stringAppend(page, "</Name>\n   <URL>");

                addLink(page, BOOL_YES);

                char pos[8];
                sprintf(pos, "%d",position);

                if (idx >= 0) {
                    stringAppend(page, XML_LIST_MENU2);
                    char buf[8];
                    sprintf(buf, "%d,",idx);
                    stringAppend(page, buf);
                } else {
                    stringAppend(page, XML_SHORT_MENU);
                }

                stringAppend(page, pos);
                stringAppend(page, "</URL>\n  </MenuItem>\n");
            }
            list = listSingleNext(list);
            position++;
        }
    }

    stringAppend(page, "  <SoftKeyItem>\n    <Name>Select</Name>\n    <URL>SoftKey:Select</URL>\n    <Position>1</Position>\n  </SoftKeyItem>\n");

    stringAppend(page, XML_LIST99);

    sendForm(fd, httpPageH, page);

    stringFree(page,      BOOL_YES);
    stringFree(httpPageH, BOOL_YES);
}

static void renderControlForm(int fd)
{
    INFO2("[WS]: renderControlForm");

    string_t* httpPageH = renderFormHead(CF, cf_caption, cf_fg, cf_bg);

    string_t* page = stringNew("<");
    stringAppend(page, cf_font->str);
    stringAppend(page, "\n<p style=\"width:");
    char buf[8];
    stringAppend(page, i2string(screenSize,buf));
    stringAppend(page, "px; overflow:auto;\">");
    stringAppend(page, cf_status->str);
    stringAppend(page, "</p>");
    stringAppend(page, "</");
    stringAppend(page, cf_font->str);


    if (bottomlineSkin && cf_cover != NULL) {
        stringAppend(page, "\n<img src=\"");
        stringAppend(page,cf_cover->str);
        stringAppend(page, "\" border=\"1\">");
    }

    stringAppend(page, HTTP_FORM);

    stringAppend(page, HTTP_TABLE1);
    stringAppend(page, i2string(iconPadding,buf));
    stringAppend(page, HTTP_TABLE2);

    int i = 0;
    int col = 1;
    while (i < 12) {

        if (bottomlineSkin && strcmp(cf_icons[i]->str,"none") == 0) {
            break; // skip if empty icon specified
        }

        stringAppend(page, HTTP_TABLE3);

        if (strcmp(cf_icons[i]->str,"none") != 0) {

            char maze[8];

            //<button accesskey="XX" type="submit" value="XX"> <img src="XX" alt="" style="vertical-align: middle">label_empty</button>
            stringAppend(page, "<button name=\"action\" accesskey=\"");
            stringAppend(page, getAccessKey(i+1));
            stringAppend(page, "\" type=\"submit\" value=\"");
            stringAppend(page, i2string(i+1,maze));
            stringAppend(page, "\"> <img src=\"");
            stringAppend(page, cf_icons[i]->str);
            stringAppend(page, ".png\" border=\"0\"");
            stringAppend(page, " alt=\"\" style=\"vertical-align: middle\">");
            stringAppend(page, "</button>");

        }

        stringAppend(page, "</td>\n");
        if (!bottomlineSkin && col == 3) {
            stringAppend(page, "    \n</tr>\n<tr>\n");
            col = 0;
        }
        i++;
        col++;
    }

    stringAppend(page, "    </tr>\n</tbody>\n</table>");
    stringAppend(page, "    </form>\n");

    string_t* httpPageT = stringNew(page->str);

    stringAppend(httpPageT, "<");
    stringAppend(httpPageT, cf_font->str);
    stringAppend(httpPageT, "\n<p style=\"width:");
    stringAppend(httpPageT, i2string(screenSize,buf));
    stringAppend(httpPageT, "px; overflow:auto;\">\n");
    stringAppend(httpPageT, cf_title->str);
    stringAppend(httpPageT, "</p>");

    if (useVolume && strlen(cf_volume->str) > 0) {
        stringAppend(httpPageT, "<p>");
        stringAppend(httpPageT, cf_volume->str);
        stringAppend(httpPageT, "%</p>");
    }
    stringAppend(httpPageT, "</");
    stringAppend(httpPageT, cf_font->str);

    addFormMenu(CF, httpPageT);

    stringAppend(httpPageT, httpTail->str);

    sendForm(fd, httpPageH, httpPageT);

    stringFree(httpPageT, BOOL_YES);
    stringFree(page,      BOOL_YES);
    stringFree(httpPageH, BOOL_YES);
}

static void renderConvertCmd(string_t* cmd)
{
    stringAppend(cmd, " -resize ");

    char num[16];
    sprintf(num,"%d",_xmlScreenW - 4);
    stringAppend(cmd, num);

    stringAppend(cmd, "x");

    sprintf(num,"%d",_xmlScreenH - 4);
    stringAppend(cmd, num);

    stringAppend(cmd, " -colorspace ");
    if (_xmlScreenG) {
        stringAppend(cmd, " Gray ");
    } else {
        stringAppend(cmd, " RGB -depth 12 ");
    }
}

static void renderXMLLayout()
{
    INFO2("[WS]:renderXMLLayout (%d,%d) pix, %s, %s", _xmlScreenW, _xmlScreenH,
          (_xmlScreenG ? "Grayscale" : "Color"),
          (cf_cover ? "with cover" : "no cover"));

    // Use ImageMagic to create picture to show

    string_t* iconpath = stringNew(confDir ? confDir : ".");
    stringAppend(iconpath,"/Icons/32/");

    string_t* cmd = NULL;

    if (bottomlineSkin) {

        int iNum = 1;
        while (iNum <= 7) {

            if (strcmp(cf_icons[iNum-1]->str,"none") == 0 || iNum == 7) {
                break; // no more icons
            }
            iNum++;
        }

        cmd = stringNew("montage -background transparent -tile ");

        char num[16];
        sprintf(num,"%dx1 -geometry 32x32 ", iNum);
        stringAppend(cmd, num);

        //stringAppend(cmd, "-draw \'text 0,0 \"");
        //stringAppend(cmd, cf_title->str);
        //stringAppend(cmd, "\";gravity Center\' ");

        int i;
        for (i=0; i<iNum; i++) {
            stringAppend(cmd, iconpath->str);
            stringAppend(cmd, cf_icons[i]->str);
            stringAppend(cmd, ".png ");
        }

        if (cf_cover) {
            stringAppend(cmd, " $HOME/.anyRemote/layout_xml_raw0.png; convert -gravity Center -background transparent ");
            stringAppend(cmd, cf_cover->str);
            stringAppend(cmd, " $HOME/.anyRemote/layout_xml_raw0.png -append $HOME/.anyRemote/layout_xml_raw.png; ");

        } else {
            stringAppend(cmd, " $HOME/.anyRemote/layout_xml_raw.png; ");
        }
        stringAppend(cmd, "convert $HOME/.anyRemote/layout_xml_raw.png ");

    } else {

        cmd = stringNew("montage -background transparent -tile 3x4 -geometry 32x32 ");   // use 32x32 icons

        //stringAppend(cmd, "-draw \'text 0,0 \"");
        //stringAppend(cmd, cf_title->str);
        //stringAppend(cmd, "\";gravity Center\' ");

        int i;
        for (i=0; i<12; i++) {
            stringAppend(cmd, iconpath->str);
            stringAppend(cmd, cf_icons[i]->str);
            stringAppend(cmd, ".png ");
        }
        stringAppend(cmd, " $HOME/.anyRemote/layout_xml_raw.png; convert $HOME/.anyRemote/layout_xml_raw.png ");
    }

    renderConvertCmd(cmd);

    stringAppend(cmd, " $HOME/.anyRemote/layout_xml.png");

    INFO2("[WS]:renderXMLLayout CMD=%s", cmd->str);

    _xmlLayoutW = _xmlScreenW; // reset
    _xmlLayoutH = _xmlScreenH;

    size_t sz = 0;
    char* dummy = executeCommandPipe(cmd->str, &sz);
    if (dummy) {

        free(dummy);

        char* buf = executeCommandPipe("identify -format \"%w,%h\" $HOME/.anyRemote/layout_xml.png", &sz);
        if (buf) {

            char* w = strtok(buf, ",");
            char* h = strtok(NULL,",");

            _xmlLayoutW = atoi(w);
            _xmlLayoutH = atoi(h);
            INFO2("[WS]:renderXMLLayout rendered image is (%d,%d)", _xmlLayoutW, _xmlLayoutH);

            free(buf);

        } else {
            ERROR2("[WS]:renderXMLLayout can not get size layout image");
        }
    } else {
        ERROR2("[WS]:renderXMLLayout can not render layout image");
    }

    stringFree(cmd,      BOOL_YES);
    stringFree(iconpath, BOOL_YES);

    _xmlLayoutOk = BOOL_YES;
}

static void renderXMLImage()
{
    INFO2("[WS]:renderXMLImage (%d,%d) pix, %s", _xmlScreenW, _xmlScreenH,
          (_xmlScreenG ? "Grayscale" : "Color"));

    // Use ImageMagic to create picture to show

    string_t* cmd = stringNew("convert ");
    stringAppend(cmd, window->str);

    renderConvertCmd(cmd);

    stringAppend(cmd, " $HOME/.anyRemote/image_xml.png");

    INFO2("[WS]:renderXMLImage CMD=%s", cmd->str);

    size_t sz = 0;
    char* str = executeCommandPipe(cmd->str, &sz);
    // it is on to get NULL in str, because command does not produce any output
    free(str);

    stringFree(cmd, BOOL_YES);
}

static void renderControlXMLForm(int fd)
{
    INFO2("[WS]: renderControlXMLForm");

    if (!_xmlLayoutOk) {
        renderXMLLayout();
    }

    string_t* httpPageH = renderXMLFormHead(CF, cf_caption);

    string_t* httpPageT = stringNew("");
    char num[16];

    stringAppend(httpPageT, "  <LocationX>");
    sprintf(num,"%d", (_xmlScreenW - _xmlLayoutW)/2);
    stringAppend(httpPageT, num);
    stringAppend(httpPageT, "</LocationX>\n");

    stringAppend(httpPageT, "  <LocationY>");
    sprintf(num,"%d", (bottomlineSkin ? _xmlScreenH - _xmlLayoutH : (_xmlScreenH - _xmlLayoutH)/2));
    stringAppend(httpPageT, num);
    stringAppend(httpPageT, "</LocationY>\n");

    stringAppend(httpPageT, "  <URL>");

    addLink(httpPageT, BOOL_YES);

    stringAppend(httpPageT, "xml_layout</URL>\n");

    int i = 0;
    while (i < 12) {

        if (bottomlineSkin && strcmp(cf_icons[i]->str,"none") == 0) {
            break; // skip if empty icon specified
        }

        stringAppend(httpPageT, "  <MenuItem>\n   <Name>");
        stringAppend(httpPageT, cf_icons[i]->str);
        stringAppend(httpPageT, "</Name>\n   <URL>");

        addLink(httpPageT, BOOL_YES);

        stringAppend(httpPageT, XML_BUTTON_PRESS);

        sprintf(num,"%d", i+1);
        stringAppend(httpPageT, num);

        stringAppend(httpPageT, "</URL>\n  </MenuItem>\n");

        i++;
    }

    addXMLFormMenu(CF, httpPageT);

    stringAppend(httpPageT, XML_CFORM99);

    sendForm(fd, httpPageH, httpPageT);

    stringFree(httpPageT, BOOL_YES);
    stringFree(httpPageH, BOOL_YES);
}

static void renderTextForm(int fd)
{
    string_t* httpPageH = renderFormHead(TX,tx_caption,tx_fg,tx_bg);

    string_t* httpPageT = stringNew("<");
    stringAppend(httpPageT, tx_font->str);
    stringAppend(httpPageT, tx_text->str);
    stringAppend(httpPageT, "</");
    stringAppend(httpPageT, tx_font->str);

    addFormMenu(TX, httpPageT);

    stringAppend(httpPageT, httpTail->str);

    sendForm(fd, httpPageH, httpPageT);

    stringFree(httpPageT, BOOL_YES);
    stringFree(httpPageH, BOOL_YES);
}

static void renderTextXMLForm(int fd)
{
    string_t* httpPageH = renderXMLFormHead(TX,tx_caption);

    string_t* httpPageT = stringNew("  <Text>");
    stringAppend(httpPageT, tx_text->str);
    stringAppend(httpPageT, "</Text>");

    //string_t* httpPageT = stringNew(tx_text->str);

    addXMLFormMenu(TX, httpPageT);

    stringAppend(httpPageT, XML_TEXT99);

    sendForm(fd, httpPageH, httpPageT);

    stringFree(httpPageT, BOOL_YES);
    stringFree(httpPageH, BOOL_YES);
}

static void renderListForm(int fd)
{
    string_t* httpPageH = renderFormHead(LI,ls_caption,ls_fg,ls_bg);

    string_t* page = stringNew("<");
    stringAppend(page, ls_font->str);
    stringAppend(page, HTTP_FORM);
    stringAppend(page, HTTP_LIST1);

    int idx = 0;
    SingleList* list = ls_items;
    char num[16];

    while (list) {

        sprintf(num,"%d", idx);

        stringAppend(page, "<option value=\"");
        stringAppend(page, num);
        stringAppend(page, "\"");
        if (idx == ls_selidx - 1) {
            stringAppend(page, " selected ");
        }
        stringAppend(page, ">");
        stringAppend(page, ((listItem*) list->data)->string->str);
        stringAppend(page, "</option>\n");

        list = listSingleNext(list);
        idx++;
    }

    string_t* httpPageT = stringNew(page->str);
    stringAppend(httpPageT, "</select>\n<hr>");

    addFormMenu(LI, httpPageT);

    stringAppend(httpPageT, "</form>\n</");
    stringAppend(httpPageT, ls_font->str);
    stringAppend(httpPageT, "\n");
    stringAppend(httpPageT, httpTail->str);

    sendForm(fd, httpPageH, httpPageT);

    stringFree(httpPageT, BOOL_YES);
    stringFree(page,      BOOL_YES);
    stringFree(httpPageH, BOOL_YES);
}

static void renderListXMLForm(int fd)
{
    string_t* httpPageH = renderXMLFormHead(LI,ls_caption);

    string_t* page = stringNew("");

    SingleList* list = ls_items;
    char num[16];

    int lsz = listSingleLength(list);


    if (lsz > 0) {

        int idx = 0;

        while (list && idx <= XML_MAX_LIST_SIZE) {

            stringAppend(page, "  <MenuItem>\n   <Name>");
            if (((listItem*) list->data)->string->len > CMXML_MAX_STRING_SIZE) {
                stringAppendLen(page, ((listItem*) list->data)->string->str, CMXML_MAX_STRING_SIZE);
            } else {
                stringAppend(page, ((listItem*) list->data)->string->str);
            }
            stringAppend(page, "</Name>\n   <URL>");

            sprintf(num,"%d", idx);

            stringAppend(page, "QueryStringParam:");
            stringAppend(page, num);
            stringAppend(page, "</URL>\n  </MenuItem>\n");

            list = listSingleNext(list);
            idx++;
        }
        if (list && idx == XML_MAX_LIST_SIZE) {
            INFO2("[WS]: renderListXMLForm: too many items in list, skip the rest");
        }

    } else {  // add dummy item
        stringAppend(page, "  <MenuItem>\n	<Name>---empty---</Name>\n   <URL>");

        stringAppend(page, "QueryStringParam:");
        stringAppend(page, "-1");
        stringAppend(page, "</URL>\n  </MenuItem>\n");
    }

    SingleList* mlist = user_menu;

    int llen = listSingleLength(mlist);

    if (llen > 0) {

        int position = 1;
        boolean_t longMenu = (llen > XML_SOFTKEY_NUM ? BOOL_YES : BOOL_NO);

        while(mlist) {

            if (mlist->data && ((string_t*) mlist->data)->str) {

                INFO2("[WS]: renderListXMLForm menu item %s",((string_t*) mlist->data)->str);

                stringAppend(page, "  <SoftKeyItem>\n    <Name>");
                if (longMenu && position == XML_SOFTKEY_NUM) {
                    stringAppend(page, ">>");
                } else {
                    stringAppend(page, ((string_t*) mlist->data)->str);
                }
                stringAppend(page, "</Name>\n    <URL>");

                addLink(page, BOOL_YES);

                char buf[8];
                sprintf(buf,"%d", position);

                if (longMenu && position == XML_SOFTKEY_NUM) {
                    stringAppend(page, XML_LIST_MENU_EXT);
                } else {
                    stringAppend(page, XML_LIST_MENU);
                    stringAppend(page, buf);
                }

                stringAppend(page, "</URL>\n    <Position>");
                stringAppend(page, buf);

                stringAppend(page, "</Position>\n  </SoftKeyItem>\n");
            }
            mlist = listSingleNext(mlist);

            position++;
            if (position > XML_SOFTKEY_NUM) {  // should not happen
                break;
            }
        }
    }

    stringAppend(page, XML_LIST99);

    sendForm(fd, httpPageH, page);

    stringFree(page,      BOOL_YES);
    stringFree(httpPageH, BOOL_YES);
}

static void renderWmanForm(int fd)
{
    INFO2("[WS]: renderWmanForm");

    string_t* httpPageH = renderFormHead(WM, cf_caption, cf_fg, cf_bg);

    string_t* httpPageT = stringNew("<");
    stringAppend(httpPageT, cf_font->str);
    stringAppend(httpPageT, cf_title->str);
    stringAppend(httpPageT, "</");
    stringAppend(httpPageT, cf_font->str);

    if (window != NULL) {
        stringAppend(httpPageT, "\n<img src=\"");
        stringAppend(httpPageT,window->str);
        stringAppend(httpPageT, "\" border=\"1\">");
    }

    addFormMenu(WM, httpPageT);

    stringAppend(httpPageT, httpTail->str);

    sendForm(fd, httpPageH, httpPageT);

    stringFree(httpPageT, BOOL_YES);
    stringFree(httpPageH, BOOL_YES);
}

static void renderWmanXMLForm(int fd)
{
    INFO2("[WS]: renderWmanXMLForm");

    string_t* httpPageH = renderXMLFormHead(WM, cf_caption);

    string_t* httpPageT = stringNew("");
    stringAppend(httpPageT, "  <LocationX>0</LocationX>\n");
    stringAppend(httpPageT, "  <LocationY>0</LocationY>\n");
    stringAppend(httpPageT, "  <URL>");

    addLink(httpPageT, BOOL_YES);

    stringAppend(httpPageT, "xml_image</URL>\n");

    addXMLFormMenu(WM, httpPageT);

    stringAppend(httpPageT, XML_WMAN99);

    sendForm(fd, httpPageH, httpPageT);

    stringFree(httpPageT, BOOL_YES);
    stringFree(httpPageH, BOOL_YES);
}

static void renderEditForm(int fd)
{
    INFO2("[WS]: renderEditForm askpass=%d", askPassword);

    string_t* httpPageH = renderFormHead(EF, ef_caption, cf_fg, cf_bg);

    string_t* httpPageT = stringNew("<");
    stringAppend(httpPageT, cf_font->str);
    stringAppend(httpPageT, HTTP_EDIT1);
    stringAppend(httpPageT, ef_label->str);

    if (askPassword) {
        stringAppend(httpPageT, HTTP_EDIT2);
    } else {
        stringAppend(httpPageT, HTTP_EDIT3);
    }

    stringAppend(httpPageT, ef_text->str);
    stringAppend(httpPageT, HTTP_EDIT4);

    addFormMenu(EF, httpPageT);

    stringAppend(httpPageT, "</form>\n</");
    stringAppend(httpPageT, cf_font->str);
    stringAppend(httpPageT, "\n");
    stringAppend(httpPageT, httpTail->str);

    sendForm(fd, httpPageH, httpPageT);

    stringFree(httpPageT, BOOL_YES);
    stringFree(httpPageH, BOOL_YES);

}

static void renderEditXMLForm(int fd)
{
    INFO2("[WS]: renderXMLEditForm askpass=%d", askPassword);

    string_t* httpPageH = renderXMLFormHead(EF, ef_caption);


    string_t* httpPageT = stringNew("  <URL>");

    addLink(httpPageT, BOOL_YES);
    stringAppend(httpPageT, ";");

    stringAppend(httpPageT, "</URL>\n  <InputItem>\n   <DisplayName>");
    stringAppend(httpPageT, ef_label->str);
    stringAppend(httpPageT, "</DisplayName>\n   <QueryStringParam>xml_ef</QueryStringParam>\n   <DefaultValue>");
    stringAppend(httpPageT, ef_text->str);
    stringAppend(httpPageT, "</DefaultValue>\n   <InputFlags>A</InputFlags>\n  </InputItem>\n");

    //addXMLFormMenu(EF, httpPageT);  -- add hardcoded Ok and Back menu items

    stringAppend(httpPageT, "  <SoftKeyItem>\n    <Name>Ok</Name>\n    <URL>SoftKey:Submit</URL>\n    <Position>1</Position>\n  </SoftKeyItem>\n");
    stringAppend(httpPageT, "  <SoftKeyItem>\n    <Name>Back</Name>\n    <URL>");
    addLink(httpPageT, BOOL_YES);
    stringAppend(httpPageT, XML_EFIELD_CANCEL);
    stringAppend(httpPageT, "</URL>\n    <Position>2</Position>\n  </SoftKeyItem>\n");
    stringAppend(httpPageT, "  <SoftKeyItem>\n    <Name>&lt;&lt;</Name>\n    <URL>SoftKey:&lt;&lt;</URL>\n    <Position>3</Position>\n  </SoftKeyItem>\n");

    stringAppend(httpPageT, XML_EFORM99);

    sendForm(fd, httpPageH, httpPageT);

    stringFree(httpPageT, BOOL_YES);
    stringFree(httpPageH, BOOL_YES);
}

static void renderCurForm(int client, int fd)
{
    int f = curForm;
    INFO2("[WS]: (%d) renderCurForm %d", client, curForm);

    switch (f) {

    case TX:
        renderTextForm(fd);
        return;

    case LI:
        renderListForm(fd);
        return;

    case EF:
        renderEditForm(fd);
        return;

    case FM:
        return;

    case WM:
        renderWmanForm(fd);
        return;

        //case CF:
    default:
        renderControlForm(fd);
        return;
    }
}

static void renderCurXMLForm(int client, int fd)
{
    int f = curForm;
    INFO2("[WS]: (%d) renderCurXMLForm %d", client, curForm);

    switch (f) {

    case TX:
        renderTextXMLForm(fd);
        return;

    case LI:
        renderListXMLForm(fd);
        return;

    case EF:
        renderEditXMLForm(fd);
        return;

    case FM:
        return;

    case WM:
        renderWmanXMLForm(fd);
        return;

        //case CF:
    default:
        renderControlXMLForm(fd);
        return;
    }
}

long parseCookie(char* buffer)
{
    //DEBUG2("[WS]: parseCookie line: %s", buffer);
    if (!buffer) {
        return 0;
    }

    char* start = buffer;
    char* startc = NULL;

    while ((start = strstr(start, "anyremote_id="))) {
        startc = start;
        start += 13;  // size("anyremote_id=") == 12

    }
    if (!startc) {
        return 0;
    }
    startc += 13;

    //DEBUG2("[WS]: parseCookie start: %s", startc);

    start = startc;
    while (startc && isdigit(*startc)) {
        startc++;
    }
    *startc = '\0';
    //DEBUG2("[WS]: parseCookie number: %s", start);

    return atol(start);
}

void parseScreenDef(char* buffer)
{
    // x-CiscoIPPhoneDisplay: 298,144,3,G
    //DEBUG2("[WS]:parseScreenDef  line: %s", buffer);

    int gap = strlen("x-CiscoIPPhoneDisplay: ");
    if (!buffer || strlen(buffer) < gap) {
        return;
    }

    char* start = buffer+gap;
    char* scrDef = strdup(start);

    char* w = strtok(scrDef, ",");
    char* h = strtok(NULL,   ",");
    /*char* d = */
    strtok(NULL,   ",");
    char* g = strtok(NULL,   ",");

    int sW      = atoi(w);
    int sH      = atoi(h);

    //printf("parseScreenDef w=%d h=%d d=%s g=%s\n", sW, sH, d, g);

    boolean_t sG = (g && g[0] == 'G');  // Grayscale

    if (_xmlLayoutOk &&
            (_xmlScreenW != sW ||
             _xmlScreenH != sH ||
             _xmlScreenG != sG)) {

        _xmlLayoutOk = BOOL_NO; // need to regenerate layout image
    }

    _xmlScreenW = sW;
    _xmlScreenH = sH;
    _xmlScreenG = sG;
}

boolean_t handleMenuOnList(int midx, int lidx, int tvisits)
{
    int iLen = listSingleLength(ls_items);
    int mLen = listSingleLength(user_menu);

    if (lidx < iLen &&             //  negative list index means empty list
            midx >= 0 && midx < mLen) {

        wMessage* wm = (wMessage*) malloc(sizeof(wMessage));
        wm->button = -1;
        wm->string = stringNew("Msg:");

        SingleList *menuItem = listSingleNth(user_menu, midx);
        stringAppend(wm->string, ((string_t*) menuItem->data)->str);

        stringAppend(wm->string, "(");
        char b[32];
        stringAppend(wm->string, i2string(lidx + 1,b));
        stringAppend(wm->string, ",");

        if (lidx >= 0) {
            SingleList *ptr = listSingleNth(ls_items, lidx);
            stringAppend(wm->string, ((listItem*) ptr->data)->string->str);
            INFO2("[WS]: choosed list item (%s)", ((listItem*) ptr->data)->string->str);
        } else { 		// list was empty
            INFO2("[WS]: choosed list item in empty list");
        }

        stringAppend(wm->string, ")");

        sendToWebServer(wm);
        return BOOL_YES;
    } else {
        INFO2("[WS]: (%d) Wrong selected index in list, ignore.", tvisits);
    }
    return BOOL_NO;
}

pointer_t serveRequest(pointer_t data)
{
    char buf[4096];
    char *method = NULL;
    char *path   = NULL;
    struct stat statbuf;

    int fd = *((int*) data);
    free(data);

    int tvisits;

    INFO2("[WS]: process request handler");

    mutexLock(M_WEB);

    tvisits = ++visits;
    clients++;
    INFO2("[WS]: (%d) process request (%d)", tvisits, clients);

    mutexUnlock(M_WEB);

    /*if (answerReady == BOOL_NO) {
        INFO2("[WS]: (%d) drop request (%d)", tvisits, clients);
        sendError(fd, 304, "Please wait", NULL, "Please wait.");
    return (void*) -1;
    }*/

    int wasRead = read(fd, buf, sizeof(buf)-1);
    if (wasRead<0) {
        return (void*) -1;
    }
    buf[wasRead] = '\0';

    INFO2("[WS]: (%d) URL: %s", tvisits, buf);

    char * cookieLine = NULL;

    char* firstLine = strtok(buf, "\r");

    int rq = RQ_UNKNOWN;
    if (!firstLine) {
        INFO2("[WS]: non-valid request");
        return (void*) -1;
    }

    if (strstr(firstLine,"POST ")) {
        rq = RQ_POST;
    } else if (strstr(firstLine,"GET ")) {
        rq = RQ_GET;
    }

    char * line       = NULL;
    char * screenDef  = NULL;
    while ((line = strtok(NULL, "\r"))) {
        if (strstr(line,"Cookie:")) {
            cookieLine = line;
        }
        if (rq == RQ_POST) {
            path = line;  // it will be last one
        }

        if (runMode == SERVER_CMXML && strstr(line,"x-CiscoIPPhoneDisplay:")) {
            screenDef = line;
        }
    }

    //parse path for GET request" GET /path HTTP/1.1
    method = strtok(firstLine, " ");
    if (rq == RQ_GET) {
        //path   = strtok(NULL, " ");
        path = firstLine + 4; // "GET "
        char *ptr = strstr(path," HTTP");
        if (ptr) {
            *ptr = '\0';
        }
    }

    long c = parseCookie(cookieLine);

    INFO2("[WS]: (%d) Path: >%s<", tvisits, path);
    INFO2("[WS]: (%d) Cookie: %ld", tvisits, c);

    if (screenDef) {
        parseScreenDef(screenDef);
    }

    char* p = NULL;

    boolean_t wait    = BOOL_NO;
    boolean_t allowed = BOOL_YES;

    if (secure == COOKIE_SENT) {

        INFO2("[WS]: (%d) secure mode, test the cookie", tvisits);

        if (c != cookie) {
            INFO2("[WS]: (%d) wrong cookie (wait for %ld)", tvisits, cookie);

            wMessage* wm = (wMessage*) malloc(sizeof(wMessage));
            wm->button = -1;
            wm->string = stringNew("_DISCONNECT_");
            sendToWebServer(wm);

            wait    = BOOL_YES;
            secure  = NEED_SEND_COOKIE;
            allowed = BOOL_NO;
        } else {
            INFO2("[WS]: (%d) cookie OK (%ld)", tvisits, cookie);
        }
    }

    if (allowed) {

        if (rq == RQ_UNKNOWN) {

            ERROR2("[WS]: (%d) Method %s is not supported", tvisits, (method?method:"UNKNOWN"));
            sendError(fd, 501, "Not supported", NULL, "Method is not supported.");

        } else if (strcmp(path,"/")  == 0 ||
                    strcmp(path,"\n") == 0) {

            if (runMode == SERVER_CMXML) {
                renderCurXMLForm(tvisits,fd);
            } else {
                renderCurForm(tvisits,fd);
            }

        } else if (strcmp(path,"/favicon.ico") == 0) {

            sendFavicon(fd);

        } else if (strstr((path+1),"/")  == NULL && strstr(path,".png")) {	// icon

            sendIcon(fd, path);
            
        } else if ((p = strstr(path,".key"))) {  // button pressed, web server

            *p = '\0';
            p--;

            while (isdigit(*p)) {
                p--;
            }
            int button = ((++p == '\0') ? -1 : atoi(p)+1);

            INFO2("[WS]: (%d) Got button %d", tvisits, button);

            wMessage* wm = (wMessage*) malloc(sizeof(wMessage));
            wm->button = button;
            wm->string = (void*) NULL;
            sendToWebServer(wm);
            wait = BOOL_YES;

        } else if ((p = strstr(path,XML_BUTTON_PRESS))) {  // button pressed, cm-xml

            p += strlen(XML_BUTTON_PRESS);

            int button = atoi(p);

            INFO2("[WS]: (%d) Got button XML %d", tvisits, button);

            wMessage* wm = (wMessage*) malloc(sizeof(wMessage));
            wm->button = button;
            wm->string = (void*) NULL;
            sendToWebServer(wm);
            wait = BOOL_YES;

        } else if ((p = strstr(path,".menu"))) {  // menu in text screen

            *p = '\0';
            p--;

            while (*p != '/') {
                p--;
            }

            INFO2("[WS]: (%d) Got menu item %s", tvisits, ++p);

            wMessage* wm = (wMessage*) malloc(sizeof(wMessage));
            wm->button = -1;
            wm->string = stringNew("");
            char * ptr = p;
            char * p2 =  NULL;

            while ((p2 = strstr(ptr,"%20"))) {   // replace %20 to space
                *p2 = '\0';
                stringAppend(wm->string, ptr);
                stringAppend(wm->string, " ");

                ptr = (p2 + 3);
            }
            stringAppend(wm->string, ptr);

            sendToWebServer(wm);
            wait = BOOL_YES;

        } else if ((p = strstr(path,XML_SHORT_MENU))) {  // menu (shorter than XML_SOFTKEY_NUM items) in CMXML

            p += strlen(XML_SHORT_MENU);

            INFO2("[WS]: (%d) Got menu item %s", tvisits, p);

            int idx = atoi(p) - 1;
            int iLen = listSingleLength(user_menu);

            if (idx >= 0 && idx < iLen) {
                SingleList *menuItem = listSingleNth(user_menu, idx);
                if (menuItem) {

                    wMessage* wm = (wMessage*) malloc(sizeof(wMessage));
                    wm->button = -1;
                    wm->string = stringNew("");
                    stringAppend(wm->string, ((string_t*) menuItem->data)->str);

                    sendToWebServer(wm);
                    wait = BOOL_YES;
                } else {
                    ERROR2("[WS]: Can not get list item #%d from list having %d items", idx, iLen);
                }
            } else {
                INFO2("[WS]: (%d) Wrong selected index in list, ignore..", tvisits);
            }

        } else if ((p = strstr(path,XML_LIST_MENU))) {  // list short menu handling in CMXML

            p += strlen(XML_LIST_MENU);
            char * q = strstr(p,"?");
            *q = '\0';
            q++;

            INFO2("[WS]: (%d) Got list menu item (%s,%s)", tvisits, p, q);

            int menu_idx = atoi(p) - 1;
            int list_idx = atoi(q);

            if (handleMenuOnList(menu_idx, list_idx, tvisits)) {
                wait = BOOL_YES;
            }

        } else if ((p = strstr(path,XML_LIST_MENU2))) {  // list menu handling (2nd step) in CMXML

            p += strlen(XML_LIST_MENU2);

            INFO2("[WS]: (%d) Got list menu item %s", tvisits, p);

            char * item = strstr(p,",");
            if (!item) {
                INFO2("[WS]: (%d) Improper item in list, ignore", tvisits);
            } else {

                (*item) = '\0';
                item++;

                int lidx = atoi(p);
                int midx = atoi(item) - 1;

                if (handleMenuOnList(midx, lidx, tvisits)) {
                    wait = BOOL_YES;
                }
            }

        } else if ((p = strstr(path, XML_LONG_MENU))) {  // menu (more than XML_SOFTKEY_NUM items) in CMXML

            p += strlen(XML_LONG_MENU);

            INFO2("[WS]: (%d) Got menu wrapper %s", tvisits, p);

            int form = atoi(p);

            sendXMLMenu(fd, form, -1);

        } else if ((p = strstr(path, XML_LIST_MENU_EXT))) {  // list menu extension in CMXML

            p += strlen(XML_LIST_MENU_EXT);
            p++;

            INFO2("[WS]: (%d) Got list item %s", tvisits, p);

            int listItem = atoi(p);

            sendXMLMenu(fd, LI, listItem);

        } else if ((p = strstr(path,"list="))) {  // menu in list screen, list item choosed

            char * item = strstr(path, HTTP_ACTION);

            if (!item) {
                INFO2("[WS]: (%d) No selected item in list, ignore", tvisits);
            } else {

                if (item) {
                    item += strlen(HTTP_ACTION);
                }

                char* plus = NULL;
                while((plus = strstr(item,"+"))) {
                    *plus = ' ';  // replace "+" back to spaces
                }

                char * amp = strstr(p,"&");
                *amp = '\0';
                char* index = p + 5;  // size("list=") == 10

                int idx = atoi(index);
                int iLen = listSingleLength(ls_items);

                if (idx >= 0 && idx < iLen) {

                    wMessage* wm = (wMessage*) malloc(sizeof(wMessage));
                    wm->button = -1;
                    wm->string = stringNew("Msg:");
                    stringAppend(wm->string, item);
                    stringAppend(wm->string, "(");
                    char b[32];
                    stringAppend(wm->string, i2string(idx+1,b));
                    stringAppend(wm->string, ",");

                    SingleList *ptr = listSingleNth(ls_items, idx);
                    stringAppend(wm->string, ((listItem*) ptr->data)->string->str);
                    INFO2("[WS]: choosed list item (%s)", ((listItem*) ptr->data)->string->str);

                    stringAppend(wm->string, ")");

                    sendToWebServer(wm);
                    wait = BOOL_YES;
                } else {
                    INFO2("[WS]: (%d) Wrong selected index in list, ignore...", tvisits);
                }
            }
        } else if ((p = strstr(path, HTTP_EDITFIELD))) {  // edit field

            char * item = strstr(path, HTTP_ACTION);

            if (!item) {
                INFO2("[WS]: (%d) No data in edit field, ignore", tvisits);
            } else {

                item += strlen(HTTP_ACTION);

                char* plus = NULL;
                while((plus = strstr(item,"+"))) {
                    *plus = ' ';  // replace "+" back to spaces
                }

                char * amp = strstr(p,"&");
                *amp = '\0';
                char* index = p + strlen(HTTP_EDITFIELD);

                wMessage* wm = (wMessage*) malloc(sizeof(wMessage));
                wm->button = -1;
                if (askPassword && strcmp("Ok", item) == 0) {
                    wm->string = stringNew("Msg:_PASSWORD_");
                    askPassword = BOOL_NO;
                } else {
                    wm->string = stringNew("Msg:");
                    stringAppend(wm->string, item);
                }
                stringAppend(wm->string, "(");
                stringAppend(wm->string, ",");
                stringAppend(wm->string, index);
                stringAppend(wm->string, ")");

                sendToWebServer(wm);
                wait = BOOL_YES;

            }
        } else if ((p = strstr(path,XML_EFIELD_SUBMIT))) {  // XML edit field OK

            char * item = p+strlen(XML_EFIELD_SUBMIT);
            INFO2("[WS]: (%d) Entered value %s", tvisits, item);

            wMessage* wm = (wMessage*) malloc(sizeof(wMessage));
            wm->button = -1;
            if (askPassword) {
                wm->string = stringNew("Msg:_PASSWORD_");
                askPassword = BOOL_NO;
            } else {
                wm->string = stringNew("Msg:");
                stringAppend(wm->string, "Ok");
            }
            stringAppend(wm->string, "(");
            stringAppend(wm->string, ",");
            stringAppend(wm->string, item);
            stringAppend(wm->string, ")");

            sendToWebServer(wm);
            wait = BOOL_YES;

        } else if ((p = strstr(path,XML_EFIELD_CANCEL))) {  // CMXML edit field Cancel


            wMessage* wm = (wMessage*) malloc(sizeof(wMessage));
            wm->button = -1;
            wm->string = stringNew("Msg:Back(,)");

            sendToWebServer(wm);
            wait = BOOL_YES;

        } else if ((p = strstr(path, HTTP_ACTION))) {  // menu in list screen, no list item choosed
            // or key press
            p += strlen(HTTP_ACTION);

            char* btn = p;
            while (isdigit(*btn)) {
                btn++;
            }

            int button = -1;
            if (btn != p) {
                *btn = '\0';
                button = atoi(p);
            }

            wMessage* wm = NULL;

            if (button > 0) {

                INFO2("[WS]: (%d) Got button %d", tvisits, button);

                wm = (wMessage*) malloc(sizeof(wMessage));
                wm->button = button;
                wm->string = (void*) NULL;

            } else {
                char* plus = NULL;
                while((plus = strstr(p,"+"))) {
                    *plus = ' ';  // replace "+" back to spaces
                }

                wm = (wMessage*) malloc(sizeof(wMessage));
                wm->button = -1;
                wm->string = stringNew("Msg:");
                stringAppend(wm->string, p);
                stringAppend(wm->string, "(,)");
            }

            if (wm) {
                sendToWebServer(wm);
                wait = BOOL_YES;
            }

        } else if (strstr(path,"xml_layout")) {  // ControlForm as image in CMXML

            sendGeneratedImage(fd, "layout_xml.png");

        } else if (strstr(path,"xml_image")) {   // WindowManager screen as image in CMXML

            renderXMLImage();
            sendGeneratedImage(fd, "image_xml.png");

        } else if (strstr(path,"anyremote.gui")) {   // iViewer GUI file reguest
        
            sendIViewerGui(fd);
       
        } else if (strstr(path,"generated_cover")) {	// cover for iViewer
            
            sendGeneratedImage(fd, "generated_cover");
     
        } else { // treat as file/dir name

            string_t* fpath = stringNew("");

            if (getIViewer()) {   // search images in $(CfgData)/Icons/iViewer

                stringAppend(fpath, confDir);
                stringAppend(fpath,"/Utils/iViewer/");
            }

            stringAppend(fpath, path);

            if (stat(fpath->str, &statbuf) < 0) {
                ERROR2("[WS]: (%d) File %s not found", tvisits, fpath->str);
                sendError(fd, 404, "Not Found", NULL, "File not found.");

                /*} else if (S_ISDIR(statbuf.st_mode)) {

                	//INFO2("[WS]: (%d)  Directory listing requested", tvisits);

                	len = fpath->len;
                	if (len == 0 || fpath->str[len - 1] != '/') {

                	    snprintf(pathbuf, sizeof(pathbuf), "Location: %s/", fpath->str);
                	    sendError(fd, 302, "Found", pathbuf, "Directories must end with a slash.");

                	} else {
                	    snprintf(pathbuf, sizeof(pathbuf), "%sindex.html", path);
                	    if (stat(pathbuf, &statbuf) >= 0) {

                        	sendFile(fd, pathbuf, &statbuf);

                	    } else {
                        	DIR *dir;
                        	struct dirent *de;

                        	sendHeaders(fd, 200, "OK", NULL, "text/html", -1, statbuf.st_mtime, 0);
                        	sprintf(f, "<HTML><HEAD><TITLE>Index of %s</TITLE></HEAD>\r\n<BODY>", fpath->str);
                    	bytes_sent = send(fd,f,strlen(f),0);

                        	sprintf(f, "<H4>Index of %s</H4>\r\n<PRE>\n", path);
                    	sendData(fd, f);

                    	sendData(fd, "Name Last Modified Size\r\n");
                    	sendData(fd, "<HR>\r\n");

                        	//if (len > 1) {
                    	//    sprintf(f, "<A HREF=\"..\">..</A>\r\n");
                        	//    sendData(fd, f);
                    	//}
                	//INFO2("[Thread] Parse %s\n", path);

                        	dir = opendir(fpath->str);
                        	while ((de = readdir(dir)) != NULL) {
                        	    char timebuf[32];
                        	    struct tm *tm;
                        	     strcpy(pathbuf, path);
                        	    strcat(pathbuf, de->d_name);

                        	    stat(pathbuf, &statbuf);
                        	    tm = gmtime(&statbuf.st_mtime);
                        	    strftime(timebuf, sizeof(timebuf), "%d-%b-%Y %H:%M:%S", tm);

                        	    sprintf(f, "<A HREF=\"%s%s\">", de->d_name, S_ISDIR(statbuf.st_mode) ? "/" : "");
                        	    sendData(fd, f);

                	    sprintf(f, "%s%s", de->d_name, S_ISDIR(statbuf.st_mode) ? "/</A>" : "</A> ");
                        	    sendData(fd, f);

                        	    if (_D_EXACT_NAMLEN(de) < 32) {
                        	sprintf(f, "%*s", (int) (32 - _D_EXACT_NAMLEN(de)), "");
                                	sendData(fd, f);
                        	    }
                        	    if (S_ISDIR(statbuf.st_mode)) {
                                	sprintf(f, "%s\r\n", timebuf);
                                	sendData(fd, f);
                        	    } else {
                                	sprintf(f, "%s %10d\r\n", timebuf, (int) statbuf.st_size);
                                	sendData(fd, f);
                        	    }
                        	}
                        	closedir(dir);

                        	sprintf(f, "</PRE>\r\n<HR>\r\n<ADDRESS>%s</ADDRESS>\r\n</BODY></HTML>\r\n", SERVER);
                	    }
                	}*/

            } else {

                INFO2("[WS] (%d) Send file", tvisits);
                sendFile(fd, fpath->str, &statbuf);
            }

            stringFree(fpath, BOOL_YES);
        }
    }

    if (wait) {
        mutexLock(M_WEB);
        answerReady = BOOL_NO;
        mutexUnlock(M_WEB);

        int waitingTime = 0;

        while (!answerReady && waitingTime < 1500) {	// 1/50*1500 = 30 sec
            // Main loop timer (1/50 of second)
            usleep(20000);
            waitingTime++;
        }

        INFO2("[WS]: (%d) Wait answer for %d (1/50 sec)", tvisits, waitingTime);

        if (runMode == SERVER_CMXML) {
            renderCurXMLForm(tvisits,fd);
        } else {
            renderCurForm(tvisits,fd);
        }
    }

    close(fd);

    mutexLock(M_WEB);
    //INFO2("[WS]: (%d) clientSockets before remove #%d", tvisits, listSingleLength(clientSockets));
    clientSockets = listSingleRemove(clientSockets, CAST_INT_TO_POINTER(fd));
    //INFO2("[WS]: (%d) clientSockets after remove #%d", tvisits, listSingleLength(clientSockets));
    clients--;
    mutexUnlock(M_WEB);

    INFO2("[WS]: (%d) Request processed", tvisits);
    threadExit(T_MAX);

    return (void*)0;
}

static pointer_t webServerMain(pointer_t data)
{
    queueNew(Q_WEB);

    confDir = getCfgDir();

    char* v1 = dupVarValue("ScreenSize");
    if (v1 != NULL) {
        screenSize = atoi(v1);
        if (screenSize <= 0) {
            screenSize = 240;
        }
        free(v1);
    }
    INFO2("[WS]: $(ScreenSize) = %d", screenSize);

    char* v2 = dupVarValue("RefreshPage");
    if (v2 != NULL) {
        refreshPage = atoi(v2);
        if (refreshPage <= 0) {
            refreshPage = -1;
        }
        free(v2);
    }

    INFO2("[WS]: $(RefreshPage) = %d", refreshPage);

    //if (runMode == SERVER_CMXML) {  iViewer also use this
        char* v3 = dupVarValue("IpAddr");
        if (v3 != NULL) {
            INFO2("[WS]: $(IpAddr) = %s", v3);
            serverIP = stringNew(v3);
            free(v3);
        } else { // try to get it ourselves
            string_t* cmd = stringNew("/sbin/ifconfig|grep 'inet addr'|cut -f 2 -d ':'|cut -f 1 -d ' '|grep -v 127.0.0.1|head -1");

            size_t sz = 0;
            char* buf = executeCommandPipe(cmd->str, &sz);
            if (buf) {
                if (*(buf+sz-1) == '\n') {
                    *(buf+sz-1) = '\0';   // strip new-line
                }
                serverIP = stringNew(buf);

                free(buf);
            }
        }
        INFO2("[WS]: set Server IP as = %s", (serverIP ? serverIP->str : "not determined"));
    //}

    struct sockaddr_in sin;
    memset((void *) &sin, 0, sizeof(sin));

    sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

    if (sock < 0) {
        logger(L_ERR, "[WS]: webServerMain: Failed to create a socket");

        threadExit(T_WEB);
        return (void*)0;
    }

    int optval = 1;
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval));

    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(port);
    int ret = bind(sock, (struct sockaddr *) &sin, sizeof(sin));
    if (ret < 0) {
        logger(L_ERR, "[WS]: webServerMain: Failed to bind a socket");
        printf("ERROR: on binding %d->%s\n", errno, strerror(errno));

        gotExitSignal = 1;

        threadExit(T_WEB);
        return (void*)0;
    }

    listen(sock, 100);
    INFO2("[WS]: webServerMain: HTTP server listening on port %d", port);

    while (runWeb) {

        logger(L_INF, "[WS]: webServerMain: HTTP server ready to accept connection");

        int s = accept(sock, NULL, NULL);
        if (s < 0) {
            errnoDebug("[WS]: accept() ",errno);  // testing debug
            if (errno == EAGAIN) {
                continue;
            } else {
                break;
            }
        }
        INFO2("[WS]: webServerMain: HTTP server accepts connection %d", s);

        int* ptr = malloc(sizeof(int));
        *ptr = s;

        mutexLock(M_WEB);
        clientSockets = listSingleAppend(clientSockets, CAST_INT_TO_POINTER(s));
        mutexUnlock(M_WEB);

        int rq = threadNew(T_MAX, serveRequest, (void *) ptr, DETACHED);
        if (rq != RC_OK) {
            logger(L_ERR, "[WS]: webServerMain: Can not run processing thread");
        }
    }

    logger(L_ERR, "[WS]: webServerMain: exit thread");
    if (sock != -1) {
        close(sock);
        sock = -1;
    }

    threadExit(T_WEB);
    return (void*)0;
}

int openWebPort(int p, int type)
{
    if (threadExists(T_WEB) == RC_OK) {
        INFO2("[WS]: openWebPort, skip initialization");
        return 0;
    }
    INFO2("[WS]: openWebPort");

    port    = p;
    runMode = type;

    mutexNew(M_WEB);

    // default control form
    cf_caption = stringNew("anyRemote");
    cf_title  = stringNew(" ");
    cf_status = stringNew(" ");
    cf_volume = stringNew("");
    cf_font   = stringNew("h6>");
    cf_fg     = stringNew("FFFFFF");
    cf_bg     = stringNew("000000");
    cf_cover  = NULL;


    // default text form
    tx_caption = stringNew("");
    tx_font    = stringNew("h6>");
    tx_fg      = stringNew("000000");
    tx_bg      = stringNew("FFFFFF");
    tx_text    = stringNew("");

    // default list form
    ls_caption = stringNew("");
    ls_font    = stringNew("h4>");
    ls_fg      = stringNew("000000");
    ls_bg      = stringNew("FFFFFF");
    ls_selidx  = 1;

    // default winndow form
    window     = NULL;

    // default edit field form
    ef_caption = stringNew("");
    ef_text    = stringNew("");
    ef_label   = stringNew("");

    httpTail   = stringNew("\n<h6><a href=\"/\">Redraw</a></h6>\n</center></body>\n</html>\n");

    int i;
    for (i=0; i<12; i++) {
        cf_icons[i] = stringNew("default");
    }

    threadNew(T_WEB,webServerMain,NULL,JOINABLE);

    return 0;
}

int openWeb(ConnectInfo* connInfo)
{
    DEBUG2("[DS]: Web Server mode. Use port %d", connInfo->port);
    if (openWebPort(connInfo->port, connInfo->mode) < 0) {
        connInfo->connected = BOOL_NO;
        return EXIT_NOK;
    }
    connInfo->connected = BOOL_YES;
    return EXIT_OK;
}

int checkWebPort(char* buf, int capacity)
{
    //logger(L_DBG,"[DS]: checkWebPort");

    if (queueExists(Q_WEB) != RC_OK) {
        return 0;
    }

    //logger(L_DBG,"[WS]: checkWebPort queueExists");

    int l = 0;

    // Verify commands from queue (timeout about 1/2 sec)
    wMessage* wm = (wMessage*) queuePop(Q_WEB);
    if (wm != NULL) {
        logger(L_DBG, "[DS]: Got event");

        if (wm->button > 0) {
            INFO2("[DS]: Button pressed %d", wm->button);
            char bbuf[16];
            if (wm->button == 10) {
                sprintf(bbuf,"Msg:%c",'*');
            } else if (wm->button == 11) {
                sprintf(bbuf,"Msg:%c",'0');
            } else if (wm->button == 12) {
                sprintf(bbuf,"Msg:%c",'#');
            } else {
                sprintf(bbuf,"Msg:%d",wm->button);
            }
            strncpy(buf,bbuf,capacity);
            l = strlen(buf);
        } else if (wm->string) {
            INFO2("[DS]: Data %s", wm->string->str);

            if (strcmp(wm->string->str,"_DISCONNECT_") == 0) {   // got wrong cookie, make disconnect
                buf[0] = '\0';
                l = EOF;
            } else {
                strncpy(buf,wm->string->str,capacity);
                l = strlen(wm->string->str);
            }
        }
        freeWMessage(wm);
    }
    return l;
}

static void freeCover()
{
    if (cf_cover) {
        stringFree(cf_cover, BOOL_YES);
        cf_cover = NULL;
    }
}

static void freeWindow()
{
    if (window) {
        stringFree(window, BOOL_YES);
        window = NULL;
    }
}

void closeWebPort(ConnectInfo* connInfo, int final)
{
    INFO2("[WS]: closeWebPort %d", final);

    // close all sockets
    if (sock != -1) {
        close(sock);
        sock = -1;
    }

    mutexLock(M_WEB);

    //INFO2("[WS]: clientSockets before cleanup #%d", listSingleLength(clientSockets));

    SingleList* list = clientSockets;
    while (list) {

        int clientFD = CAST_POINTER_TO_INT(list->data);
        if (clientFD > 0) {
            close(clientFD);
        }
        list = listSingleNext(list);
    }

    listSingleFree(clientSockets);
    clientSockets = NULL;

    mutexUnlock(M_WEB);

    if (final) {
        runWeb = BOOL_NO;

        stringFree(cf_caption,BOOL_YES);
        stringFree(cf_title,  BOOL_YES);
        stringFree(cf_status, BOOL_YES);
        stringFree(cf_volume, BOOL_YES);
        stringFree(cf_font,   BOOL_YES);
        stringFree(cf_fg,     BOOL_YES);
        stringFree(cf_bg,     BOOL_YES);
        freeCover();

        stringFree(tx_caption,BOOL_YES);
        stringFree(tx_font,   BOOL_YES);
        stringFree(tx_fg,     BOOL_YES);
        stringFree(tx_bg,     BOOL_YES);
        stringFree(tx_text,   BOOL_YES);

        stringFree(ls_caption,BOOL_YES);
        stringFree(ls_font,   BOOL_YES);
        stringFree(ls_fg,     BOOL_YES);
        stringFree(ls_bg,     BOOL_YES);
        freeList();

        freeWindow();

        stringFree(ef_caption,BOOL_YES);
        stringFree(ef_text,   BOOL_YES);
        stringFree(ef_label,  BOOL_YES);

        freeMenu();

        stringFree(httpTail,   BOOL_YES);

        if (serverIP) {
            stringFree(serverIP, BOOL_YES);
        }

        int i;
        for (i=0; i<12; i++) {
            stringFree(cf_icons[i],BOOL_YES);
        }

        if (confDir) {
            free(confDir);
        }

        mutexRemove(M_WEB);
        queueRemove(Q_WEB, freeWMessage);
    }

    return;
}

static void stripCommandEnding(char *s)
{
    if (s && strlen(s) >= 2) {
        s[strlen(s)-2] = '\0';   // strip );
    }
}

static void setCaption(char * s)
{
    stringAssign(cf_caption,s);
}

static void setTitle(char * s)
{
    boolean_t setupMenu = (curForm != CF);

    stringAssign(cf_title,(s?s:" "));

    if (runMode == SERVER_CMXML) {
        stringTruncate(cf_title,CMXML_MAX_STRING_SIZE);
    }

    curForm = CF;

    if (setupMenu) {
        setupDefMenu();
    }
}

static void setStatus(char * s)
{
    boolean_t setupMenu = (curForm != CF);

    stringAssign(cf_status,(s?s:" "));
    curForm = CF;

    if (setupMenu) {
        setupDefMenu();
    }

}

static void setIcons()
{
    boolean_t setupMenu = (curForm != CF);

    char* token = strtok(NULL,",");

    if (token && strcmp(token,"SAME") != 0) {
        setCaption(token);
    }

    token = strtok(NULL,",");

    while (token) {

        while (isspace(*token)) {
            ++token;
        }

        int ic = -1;
        if (strncmp(token,"*",1) == 0) {
            ic = 9;
        } else if (strncmp(token,"#",1) == 0) {
            ic = 11;
        } else if (strncmp(token,"0",1) == 0) {
            ic = 10;
        } else {
            ic = atoi(token)-1;
        }
        if (ic >=0 && ic < 12) {
            token = strtok(NULL,",");
            while (isspace(*token)) {
                ++token;
            }

            stringAssign(cf_icons[ic],(token?token:"none"));
            INFO2("[WS]: setIcons %d %s",ic,cf_icons[ic]->str);
        }
        token = strtok(NULL,",");
    }

    curForm = CF;

    if (setupMenu) {
        setupDefMenu();
    }

    _xmlLayoutOk = BOOL_NO;  // Need to regenerate
}

//
// Set(skin,default|bottomline|3x4|7x1 <--- handle only this for now
//    [,keypad_only|joystick_only]
//    [,ticker|noticker] [,volume]
//    [,choose,_button_] [,up,_button_] [,down,_button_])
//
static void setSkin()
{
    useVolume      = BOOL_NO;  // reset volume usage flag
    bottomlineSkin = BOOL_NO;
    char* token;

    boolean_t setupMenu = (curForm != CF);

    while((token = strtok(NULL,","))) {

        if (strcmp(token,"bottomline") == 0 || strcmp(token,"7x1") == 0) {
            bottomlineSkin = BOOL_YES;
        } else if (strcmp(token,"volume") == 0) {
            useVolume = BOOL_YES;
        }
    }

    if (bottomlineSkin) {
        freeCover();
    }

    curForm = CF;

    if (setupMenu) {
        setupDefMenu();
    }

    _xmlLayoutOk = BOOL_NO;  // Need to regenerate
}

static void setCover(const char* file)
{
    freeCover();
    if (file) {
        cf_cover = stringNew(file);
    }
    _xmlLayoutOk = BOOL_NO;  // Need to regenerate
}

static void setFont(int form)
{
    char *s = strtok(NULL,",");
    INFO2("[WS]: setFont %d %s",form,(s?s:"NULL"));

    if (!s) {
        return;
    }

    while (isspace(*s)) {
        ++s;
    }

    if (form == CF) {
        if (strncmp(s,"small",5) == 0) {
            stringAssign(cf_font,"h6>");
        } else if (strncmp(s,"medium",6) == 0) {
            stringAssign(cf_font,"h4>");
        } else if (strncmp(s,"large",5) == 0) {
            stringAssign(cf_font,"h2>");
        }
    } else if (form == TX) {
        if (strncmp(s,"small",5) == 0) {
            stringAssign(tx_font,"h6>");
        } else if (strncmp(s,"medium",6) == 0) {
            stringAssign(tx_font,"h4>");
        } else if (strncmp(s,"large",5) == 0) {
            stringAssign(tx_font,"h2>");
        }
    } else if (form == LI) {
        if (strncmp(s,"small",5) == 0) {
            stringAssign(ls_font,"h6>");
        } else if (strncmp(s,"medium",6) == 0) {
            stringAssign(ls_font,"h4>");
        } else if (strncmp(s,"large",5) == 0) {
            stringAssign(ls_font,"h2>");
        }
    }
}

static void setFgBg(int form, boolean_t set_fg)
{
    char buf[16];

    char *token = strtok(NULL,",");
    if (!token) {
        return;
    }
    int r = atoi(token);

    token = strtok(NULL,",");
    if (!token) {
        return;
    }
    int g = atoi(token);

    token = strtok(NULL,",");
    if (!token) {
        return;
    }
    int b = atoi(token);

    sprintf(buf,"%0*x%0*x%0*x",2,r,2,g,2,b);

    INFO2("[WS]: setFgBg fg=%d form=%d color=%s",set_fg,form,buf);

    if (form == CF) {
        if (set_fg) {
            stringAssign(cf_fg,buf);
        } else {
            stringAssign(cf_bg,buf);
        }
    } else if (form == TX) {
        if (set_fg) {
            stringAssign(tx_fg,buf);
        } else {
            stringAssign(tx_bg,buf);
        }
    } else if (form == LI) {
        if (set_fg) {
            stringAssign(ls_fg,buf);
        } else {
            stringAssign(ls_bg,buf);
        }
    }
}

static void setText()
{
    //logger(L_INF, "[WS]: setText");
    boolean_t setupMenu = (curForm != TX);

    char *token = strtok(NULL,",");
    if (!token) {
        return;
    }
    while (isspace(*token)) {
        ++token;
    }

    if (strncmp(token,"add",3) == 0 ||
            strncmp(token,"replace",7) == 0) {

        curForm = TX;

        char *token2 = strtok(NULL,","); // can be NULL
        if (!token2) {
            return;
        }

        //logger(L_INF, "[WS]: setText add/replace");

        if (strncmp(token,"replace",7) == 0) {
            stringTruncate(tx_text,0);
        }

        if (strcmp(token2,"SAME") != 0) {
            stringAssign(tx_caption,token2);
        }

        char *token3 = strtok(NULL,"\n");
        while (token3) {

            //INFO2("[WS]: Set(text, ...) %s", token3);
            if (runMode == SERVER_CMXML) {
                stringAppend(tx_text, token3);
                stringAppend(tx_text, "&#10;");

            } else {
                stringAppend(tx_text, "<p style=\"width:");
                char buf[8];
                stringAppend(tx_text, i2string(screenSize,buf));
                stringAppend(tx_text, "px; overflow:auto;\">");

                stringAppend(tx_text, token3);
                stringAppend(tx_text, "</p>");
            }
            token3 = strtok(NULL,"\n");
        }
    } else if (strcmp(token,"close") == 0) {
        char *token2 = strtok(NULL,","); // can be NULL
        if (token2 && strcmp(token,"clear") == 0) {
            stringTruncate(tx_text,0);
        }
        curForm = CF;
        setupDefMenu();
        return;
    } else if (strncmp(token,"clear",5) == 0) {
        stringTruncate(tx_text,0);
        return;
    } else if (strncmp(token,"fg",2) == 0) {
        setFgBg(TX,BOOL_YES);
        return;
    } else if (strncmp(token,"bg",2) == 0) {
        setFgBg(TX,BOOL_NO);
        return;
    } else if (strncmp(token,"font",4) == 0) {
        setFont(TX);
        return;
    } else if (strncmp(token,"show",4) == 0) {
        curForm = TX;
    } else {
        ERROR2("[WS]: Can to parse Set(text, ...)");
        return;
    }
    if (setupMenu) {
        setupDefMenu();
    }
}

static void setList(boolean_t useIcons)
{
    logger(L_INF, "[WS]: setList");
    boolean_t setupMenu = (curForm != LI);

    char *token = strtok(NULL,",");
    if (!token) {
        return;
    }
    while (isspace(*token)) {
        ++token;
    }

    if (strncmp(token,"add",3) == 0 ||
            strncmp(token,"replace",7) == 0) {

        curForm = LI;

        char *token2 = strtok(NULL,","); // can be NULL
        if (!token2) {
            return;
        }

        logger(L_INF, "[WS]: setList add/replace");

        if (strcmp(token,"replace") == 0) {
            freeList();
        }

        if (strcmp(token2,"SAME") != 0) {
            stringAssign(ls_caption,token2);
        }

        //INFO2("[WS]: setList title %s", token2);

        char *token3 = strtok(NULL,",");
        while (token3) {

            if (*token3 == '\n') {        // remove new-line at the start
                token3++;
            }

            listItem * item = malloc(sizeof(listItem));

            //INFO2("[WS]: Set(list, ...) %s", token3);
            int l = strlen(token3);
            if(*(token3 + l) == '\n') {     // remove new-line at the end
                *(token3 + l) = '\0';
            }
            if(*(token3 + l - 1) == '\n') {  // shit happens
                *(token3 + l - 1) = '\0';
            }

            if (strlen(token3) > 0) {	// avoid empty list item
                char * semicolon = index(token3,':');
                if (useIcons && semicolon) {
                    *semicolon = '\0';
                    semicolon++;
                    item->icon   = stringNew(token3);
                    item->string = stringNew(semicolon);
                    //INFO2("[WS]: Set(iconlist, ...) %s : %s", token3, semicolon);
                } else {
                    item->icon   = NULL;
                    item->string = stringNew(token3);
                }

                ls_items = listSingleAppend(ls_items, item);

                //INFO2("[WS]: list size #%d", listSingleLength(ls_items));
            }

            token3 = strtok(NULL,",");
        }

    } else if (strncmp(token,"close",5) == 0) {
        char *token2 = strtok(NULL,","); // can be NULL
        if (token2 && strcmp(token,"clear") == 0) {
            freeList();
        }
        curForm = CF;
        setupDefMenu();
        return;
    } else if (strncmp(token,"clear",5) == 0) {
        freeList();
        return;
    } else if (strncmp(token,"fg",2) == 0) {
        setFgBg(LI,BOOL_YES);
        return;
    } else if (strncmp(token,"bg",2) == 0) {
        setFgBg(LI,BOOL_NO);
        return;
    } else if (strncmp(token,"font",4) == 0) {
        setFont(LI);
        return;
    } else if (strncmp(token,"select",6) == 0) {
        curForm = LI;
        char *token2 = strtok(NULL,",");
        if (token2) {
            ls_selidx = atoi(token2);
        }
    } else if (strncmp(token,"show",4) == 0) {
        curForm = LI;
    } else {
        ERROR2("[WS]: Can to parse Set([icon]list, ...)");
        return;
    }
    if (setupMenu) {
        setupDefMenu();
    }
}

static void setEditfield()
{
    logger(L_INF, "[WS]: setEditfield");

    boolean_t setupMenu = (curForm != EF);

    char *token1 = strtok(NULL,",");
    if (!token1) {
        return;
    }

    char *token2 = strtok(NULL,",");
    if (!token2) {
        return;
    }

    char *token3 = strtok(NULL,",");  // can be NULL

    stringAssign(ef_caption,token1);
    stringAssign(ef_label,  token2);
    stringAssign(ef_text,   token3 ? token3 : "");

    askPassword = BOOL_NO;
    curForm = EF;
    if (setupMenu) {
        setupDefMenu();
    }
    INFO2("[WS]: setEditfield done %d", curForm);
}

//
// Set(image,window,_image_file_name_)
// Set(image,icon,_icon_name_,_image_file_name_) -- not supported
// Set(image,show|close|cursor|nocursor|remove_all) -- only 1,2 are supported
//
static void setImage(const char* cmd, const char* file)
{
    INFO2("[WS]: setImage %s",cmd);
    boolean_t setupMenu = (curForm != WM);

    if (!cmd) {
        return;
    }
    while (isspace(*cmd)) {
        ++cmd;
    }

    if (strncmp(cmd,"window",6) == 0) {

        logger(L_INF, "[WS]: setImage: window");

        curForm = WM;

        if (!file) {
            return;
        }
        while (isspace(*file)) {
            ++file;
        }

        freeWindow();
        window = stringNew(file);

    } else if (strncmp(cmd,"show",4) == 0) {
        curForm = WM;
    } else if (strncmp(cmd,"close",5) == 0) {
        curForm = CF;
        setupDefMenu();
        return;
    } else {
        logger(L_INF, "[WS]: setImage: skip command");
        return;
    }

    if (setupMenu) {
        setupDefMenu();
    }
}

static void setMenu()
{
    char *token = strtok(NULL,",");
    if (!token) {
        return;
    }
    while (isspace(*token)) {
        ++token;
    }

    if (strncmp(token,"add",3) == 0 ||
            strncmp(token,"replace",7) == 0) {

        logger(L_INF, "[WS]: setMenu add/replace");

        if (strcmp(token,"replace") == 0) {
            setupDefMenu();
        }

        char *token3 = strtok(NULL,",");
        while (token3) {

            while (isspace(*token3)) {
                ++token3;
            }

            INFO2("[WS]: Set(menu, ...) %s", token3);

            string_t* dup = stringNew(token3);
            //INFO2("[WS]: Set(menu, ...) DUPED");
            user_menu = listSingleAppend(user_menu, dup);
            //INFO2("[WS]: Set(menu, ...) APPENDED");
            token3 = strtok(NULL,",");
        }
        //INFO2("[WS]: Set(menu, ...) PROCESSED");

    } else if (strncmp(token,"clear",5) == 0) {
        freeMenu();
    } else {
        ERROR2("[WS]: Can to parse Set(menu, ...)");
    }

}

static void setParams()
{
    char *token = strtok(NULL,",");
    if (!token) {
        return;
    }

    while (token) {

        while (isspace(*token)) {
            ++token;
        }

        if (strncmp(token,"icon_padding",12) == 0) {

            char *token2 = strtok(NULL,",");
            if (!token2) {
                return;
            }

            while (isspace(*token2)) {
                ++token2;
            }

            iconPadding = atoi(token2);
            if (iconPadding <= 0) {
                iconPadding = 0;
            }

            INFO2("[WS]: Use icon padding %d", iconPadding);

        } else if (strncmp(token,"icon_size",9) == 0) {

            char *token2 = strtok(NULL,",");
            if (!token2) {
                return;
            }

            while (isspace(*token2)) {
                ++token2;
            }

            iconSizeScale = atoi(token2);
            if (iconSizeScale <= 0) {
                iconSizeScale = 64;
            }

            if (iconSizeScale <= 32) {
                iconSize = 32;
            } else if (iconSizeScale <= 48) {
                iconSize = 48;
            } else if (iconSizeScale <= 64) {
                iconSize = 64;
            } else {
                iconSize = 128;
            }

            INFO2("[WS]: Use icon padding %d", iconPadding);

        } else {
            ERROR2("[WS]: Skip Set(parameter, ...) %s", token);
        }

        token = strtok(NULL,",");
    }
}

static void getPass()
{
    logger(L_INF, "[WS]: getPass");

    boolean_t setupMenu = (curForm != EF);

    stringAssign(ef_caption,"Enter password");
    stringAssign(ef_label,  "Enter password");
    stringAssign(ef_text,   "");

    askPassword = BOOL_YES;
    secure      = NEED_SEND_COOKIE;
    cookie      = random();

    curForm = EF;

    if (setupMenu) {
        setupDefMenu();
    }
}

int writeWebConnStr(const char* value)
{
    INFO2("[WS]: writeWebConnStr %s", value);

    wMessage* wm  = NULL;
    wMessage* wm2 = NULL;

    char * cmd = strdup(value);

    //boolean_t skipSpaces = BOOL_NO;
    //INFO2("[WS]: parse lenght %d", (cmd ? (int) strlen(cmd) : 0));

    stripCommandEnding(cmd);

    if (strlen(cmd) < MAXMAXLEN) {
        INFO2("[WS]: parse %d %s", curForm, cmd);
    } else {
        INFO2("[WS]: parse %d ... command too long ...", curForm);
    }

    char* token = strtok(cmd,",");
    while (isspace(*token)) {
        ++token;
    }

    mutexLock(M_WEB);

    if (strncmp(token,"Set(status",10) == 0) {
        setStatus(cmd+strlen("Set(status,"));
    } else if (strncmp(token,"Set(title",9) == 0) {
        setTitle(cmd+strlen("Set(title,"));
    } else if (strncmp(token,"Set(icons",9) == 0) {
        setIcons();
    } else if (strncmp(token,"Set(font",8) == 0) {
        setFont(CF);
    } else if (strncmp(token,"Set(fg",6) == 0) {
        setFgBg(CF,BOOL_YES);
    } else if (strncmp(token,"Set(bg",6) == 0) {
        setFgBg(CF,BOOL_NO);
    } else if (strncmp(token,"Set(volume",10) == 0) {
        char* sz = strtok(NULL,",");
        if (sz) {
            stringAssign(cf_volume,sz);
        } else {
            stringAssign(cf_volume,"");
        }
    } else if (strncmp(token,"Set(layout",10) == 0 ||
               strncmp(token,"Set(skin",8) == 0) {
        setSkin();
    } else if (strncmp(token,"Set(cover",9) == 0) {
        ERROR2("[WS]: Improperly formed command Set(cover,...)");
    } else if (strncmp(token,"Set(list",8) == 0) {
        setList(BOOL_NO);
    } else if (strncmp(token,"Set(iconlist",12) == 0) {
        setList(BOOL_YES);
    } else if (strncmp(token,"Set(text",8) == 0) {
        setText();
    } else if (strncmp(token,"Set(menu",8) == 0) {
        setMenu();
    } else if (strncmp(token,"Set(editfield",13) == 0) {
        setEditfield();
    } else if (strncmp(token,"Set(image",9) == 0) {
        ERROR2("[WS]: Improperly formed command Set(image,...)");
    } else if (strncmp(token,"Set(popup",9) == 0) {
        // ignore
    } else if (strncmp(token,"Set(disconnect",14) == 0) {
        // ignore
    } else if (strncmp(token,"Set(fullscreen",16) == 0) {
        // ignore
    } else if (strncmp(token,"Set(vibrate",11) == 0) {
        // ignore
    } else if (strncmp(token,"Set(parameter",13) == 0) {
        setParams();
    } else if (strncmp(token,"Get(password",12) == 0) {
        getPass();
    } else if (strncmp(token,"Get(screen_size",16) == 0) {

        char b[32];
        const char *buf = i2string(screenSize,b);

        //SizeX(width,) and SizeY(height,) - use same

        wm = (wMessage*) malloc(sizeof(wMessage));
        wm->button = -1;
        wm->string = stringNew("SizeX(");
        stringAppend(wm->string,buf);
        stringAppend(wm->string,",)");

        wm2 = (wMessage*) malloc(sizeof(wMessage));
        wm2->button = -1;
        wm2->string = stringNew("SizeY(");
        stringAppend(wm2->string,buf);
        stringAppend(wm2->string,",)");

    } else if (strncmp(token,"Get(icon_padding",16) == 0) {

        char b[32];
        const char *buf = i2string(iconPadding,b);

        wm = (wMessage*) malloc(sizeof(wMessage));
        wm->button = -1;
        wm->string = stringNew("IconPadding(");
        stringAppend(wm->string,buf);
        stringAppend(wm->string,",)");

    } else if (strncmp(token,"Get(model",9) == 0) {

        wm = (wMessage*) malloc(sizeof(wMessage));
        wm->button = -1;
        wm->string = stringNew("WebInterface");

    } else if (strncmp(token,"Get(is_exists",13) == 0) {

        char* sz = strtok(NULL,",");
        char* icon = strtok(NULL,",");
        if (sz && icon) {
            while (isspace(*sz)) {
                ++sz;
            }
            while (isspace(*icon)) {
                ++icon;
            }

            wm = (wMessage*) malloc(sizeof(wMessage));
            wm->button = -1;
            wm->string = stringNew("IconExists(");
            stringAppend(wm->string,sz);
            stringAppend(wm->string,",");
            stringAppend(wm->string,icon);
        }

    } else if (strncmp(token,"Get(cover_size",14) == 0) {

        wm = (wMessage*) malloc(sizeof(wMessage));
        wm->button = -1;
        wm->string = stringNew("CoverSize(");
        char b[32];

        stringAppend(wm->string,
                     i2string((runMode == SERVER_CMXML ? _xmlScreenH : (screenSize*8)/10),b));

        stringAppend(wm->string,",)");

    } else if (strncmp(token,"Get(version",11) == 0) {

        wm = (wMessage*) malloc(sizeof(wMessage));
        wm->button = -1;
        wm->string = stringNew("Version(,web_interface)");

    } else if (strncmp(token,"Get(cursor",10) == 0) {
        // ignore
    } else if (strncmp(token,"Get(ping",8) == 0) {

        wm = (wMessage*) malloc(sizeof(wMessage));
        wm->button = -1;
        wm->string = stringNew("Ping");

    } else if (strncmp(token,"End(",4) == 0) {
        answerReady = BOOL_YES;
    } else {
        ERROR2("[WS]: Unknown command %s", token);
    }

    mutexUnlock(M_WEB);

    if (wm) {
        sendToWebServer(wm);
    }
    if (wm2) {
        sendToWebServer(wm2);
    }

    free(cmd);
    INFO2("[WS]: parsed %d", curForm);
    return EXIT_OK;
}

int writeWebConnFile(const dMessage* dm)
{
    //INFO2("[WS]: writeWebConnFile: %s", (const char*) dm->value);

    mutexLock(M_WEB);

    if (strncmp(dm->value,"Set(cover",9) == 0) {
        setCover(dm->file);
    } else if (strncmp(dm->value,"Set(image",9) == 0) {
        setImage((dm->value + 10), dm->file);  // 10 = 9 + ","
    } else {
        ERROR2("[WS]: Unknown command file:%s", (const char*) dm->value);
    }

    mutexUnlock(M_WEB);

    INFO2("[WS]: parsed %d", curForm);
    return EXIT_OK;
}

int writeWebConn(const dMessage* dm)
{
    INFO2("[WS]: writeWebConn (%d)", dm->size);
    if (dm->type == DM_SET) {

        return writeWebConnStr(dm->value);

    } else if (dm->type == DM_SETFILE) {

        return writeWebConnFile(dm);

    } else {
        ERROR2("[WS]: Not supported");
        return EXIT_OK;
    }
}

//
// Special case for Web/CMXML servers
//
int writeWeb(ConnectInfo* peer, const dMessage* dm)
{
    // suppose peer->mode == SERVER_WEB || peer->mode == SERVER_CMXML
    if (dm->type == DM_SENDB || dm->type == DM_SENDS) {
        return EXIT_NOK;  // skip that
    }
    return writeWebConn(dm);
}

